from ._impl.dataset import BuildExecutor, MNISTBuildExecutor

# TODO: add dataset build/push/list/info api

__all__ = ["list", "push", "info", "build", "BuildExecutor", "MNISTBuildExecutor"]


def list(filter: str):
    pass


def push(swds: str):
    pass


def info(swds: str):
    pass


def build():
    pass



import os
import json
import typing as t
from pathlib import Path

import yaml
from loguru import logger

from starwhale.utils import console, now_str, is_darwin, gen_uniq_version
from starwhale.consts import (
    JSON_INDENT,
    CURRENT_FNAME,
    DataLoaderKind,
    DefaultYAMLName,
    SWDSBackendType,
    VERSION_PREFIX_CNT,
    DEFAULT_MANIFEST_NAME,
    DEFAULT_INPUT_JSON_FNAME,
    CNTR_DEFAULT_PIP_CACHE_DIR,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.base.type import URIType, EvalTaskType, RunSubDirType
from starwhale.utils.error import NoSupportError, FieldTypeOrValueError
from starwhale.utils.process import check_call
from starwhale.utils.progress import run_with_progress_bar
from starwhale.api._impl.model import PipelineHandler
from starwhale.core.model.model import StandaloneModel
from starwhale.core.dataset.model import Dataset
from starwhale.core.dataset.store import DatasetStorage
from starwhale.core.runtime.model import StandaloneRuntime

_CNTR_WORKDIR = "/opt/starwhale"
_STATUS = PipelineHandler.STATUS


# TODO: add DAG
class EvalExecutor(object):
    def __init__(
        self,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        project_uri: URI,
        name: str = "",
        desc: str = "",
        gencmd: bool = False,
        use_docker: bool = False,
    ) -> None:
        self.name = name
        self.desc = desc
        self.model_uri = model_uri
        self.project_uri = project_uri
        self.dataset_uris = [
            URI(u, expected_type=URIType.DATASET) for u in dataset_uris
        ]

        self.runtime_uri = runtime_uri

        if runtime_uri:
            self.runtime = StandaloneRuntime(
                URI(runtime_uri, expected_type=URIType.RUNTIME)
            )
            self.baseimage = self.runtime.store.get_docker_base_image()
        else:
            self.runtime = None
            self.baseimage = ""

        self.project_dir = Path(self.project_uri.real_request_uri)

        self.gencmd = gencmd
        self.use_docker = use_docker

        self._version = ""
        self._manifest: t.Dict[str, t.Any] = {"status": _STATUS.START}
        self._workdir = Path()
        self._model_dir = Path()
        self._runtime_dir = Path()

        self._do_validate()

    def _do_validate(self) -> None:
        if self.use_docker:
            if not self.runtime_uri:
                raise FieldTypeOrValueError("runtime_uri is none")
            if is_darwin(arm=True):
                raise NoSupportError(
                    "use docker as the evaluation job environment in MacOSX system (Apple Silicon processor)"
                )

    def __str__(self) -> str:
        return f"Evaluation Executor: {self.name}"

    def __repr__(self) -> str:
        return f"Evaluation Executor: name -> {self.name}, version -> {self._version}"

    def run(self, phase: str = EvalTaskType.ALL) -> str:
        try:
            self._do_run(phase)
        except Exception as e:
            self._manifest["status"] = _STATUS.FAILED
            self._manifest["error_message"] = str(e)
            raise
        finally:
            self._render_manifest()

        return self._version

    def _do_run(self, phase: str = EvalTaskType.ALL) -> None:
        self._manifest["phase"] = phase
        self._manifest["status"] = _STATUS.RUNNING

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_workdir, 5, "prepare workdir"),
            (self._extract_swmp, 15, "extract model"),
            (self._extract_swrt, 15, "extract runtime"),
            (self._gen_swds_fuse_json, 10, "gen swds fuse json"),
        ]

        _ppl = (self._do_run_ppl, 70, "run ppl")
        _cmp = (self._do_run_cmp, 70, "run cmp")

        if phase == EvalTaskType.ALL:
            operations.extend([_ppl, _cmp])
        elif phase == EvalTaskType.PPL:
            operations.append(_ppl)
        elif phase == EvalTaskType.CMP:
            operations.append(_cmp)

        run_with_progress_bar("eval run in local...", operations)

    def _gen_version(self) -> None:
        # TODO: abstract base class or mixin class for swmp/swds/
        logger.info("[step:version]create eval job version...")
        if not self._version:
            self._version = gen_uniq_version(self.name)
        self._manifest["version"] = self._version
        self._manifest["created_at"] = now_str()  # type: ignore
        logger.info(f"[step:version]eval job version is {self._version}")

    @property
    def _ppl_workdir(self) -> Path:
        return self._workdir / EvalTaskType.PPL

    @property
    def _cmp_workdir(self) -> Path:
        return self._workdir / EvalTaskType.CMP

    def _prepare_workdir(self) -> None:
        logger.info("[step:prepare]create eval workdir...")
        # TODO: fix _workdir sequence-dependency issue
        self._workdir = (
            self.project_dir
            / URIType.JOB
            / self._version[:VERSION_PREFIX_CNT]
            / self._version
        )

        ensure_dir(self._workdir)
        for _w in (self._ppl_workdir, self._cmp_workdir):
            for _n in (
                RunSubDirType.RESULT,
                RunSubDirType.DATASET,
                RunSubDirType.PPL_RESULT,
                RunSubDirType.STATUS,
                RunSubDirType.LOG,
                RunSubDirType.SWMP,
                RunSubDirType.CONFIG,
            ):
                ensure_dir(_w / _n)

        logger.info(f"[step:prepare]eval workdir: {self._workdir}")

    def _extract_swmp(self) -> None:
        _workdir = Path(self.model_uri)
        _model_yaml_path = _workdir / DefaultYAMLName.MODEL

        if _workdir.exists() and _model_yaml_path.exists() and not self.use_docker:
            self._model_dir = _workdir
        else:
            model_uri = URI(self.model_uri, expected_type=URIType.MODEL)
            _m = StandaloneModel(model_uri)
            self._model_dir = _m.extract() / "src"

    def _extract_swrt(self) -> None:
        if self.runtime and self.use_docker:
            self._runtime_dir = self.runtime.extract()
        else:
            self._runtime_dir = Path()

    def _gen_swds_fuse_json(self) -> Path:
        _fuse_jsons = []
        for _uri in self.dataset_uris:
            _store = DatasetStorage(_uri)
            fname = Dataset.render_fuse_json(_store.loc, force=False)
            _fuse_jsons.append(fname)
            logger.debug(f"[gen fuse input.json]{fname}")

        _base = json.load(open(_fuse_jsons[0], "r"))
        for _f in _fuse_jsons[1:]:
            _config = json.load(open(_f, "r"))
            _base["swds"].extend(_config["swds"])

        if self.use_docker:
            _bucket = f"{_CNTR_WORKDIR}/{RunSubDirType.DATASET}"
            for i in range(len(_base["swds"])):
                _base["swds"][i]["bucket"] = _bucket

        _json_f: Path = (
            self._workdir
            / EvalTaskType.PPL
            / RunSubDirType.CONFIG
            / DEFAULT_INPUT_JSON_FNAME
        )
        ensure_file(_json_f, json.dumps(_base, indent=JSON_INDENT))
        return _json_f

    def _gen_jsonl_fuse_json(self) -> Path:
        if self.use_docker:
            _base_dir = f"{_CNTR_WORKDIR}/{RunSubDirType.PPL_RESULT}"
        else:
            _base_dir = str(self._ppl_workdir / RunSubDirType.RESULT)

        _fuse = dict(
            backend=SWDSBackendType.FUSE,
            kind=DataLoaderKind.JSONL,
            swds=[
                dict(
                    bucket=_base_dir,
                    key=dict(
                        data=CURRENT_FNAME,
                    ),
                )
            ],
        )
        _f = (
            self._workdir
            / EvalTaskType.CMP
            / RunSubDirType.CONFIG
            / DEFAULT_INPUT_JSON_FNAME
        )
        ensure_file(_f, json.dumps(_fuse, indent=JSON_INDENT))
        return _f

    def _do_run_cmp(self) -> None:
        self._gen_jsonl_fuse_json()
        self._do_run_cmd(EvalTaskType.CMP)

    def _do_run_ppl(self) -> None:
        self._do_run_cmd(EvalTaskType.PPL)

    def _do_run_cmd(self, typ: str) -> None:
        if self.use_docker:
            self._do_run_cmd_in_container(typ)
        else:
            self._do_run_cmd_in_host(typ)

    def _do_run_cmd_in_host(self, typ: str) -> None:
        from starwhale.core.model.model import StandaloneModel

        if typ == EvalTaskType.PPL:
            _base_dir = self._ppl_workdir
        elif typ == EvalTaskType.CMP:
            _base_dir = self._cmp_workdir
        else:
            raise NoSupportError(typ)

        StandaloneModel.eval_user_handler(
            typ=typ,
            workdir=self._model_dir,
            kw={
                "status_dir": _base_dir / RunSubDirType.STATUS,
                "log_dir": _base_dir / RunSubDirType.LOG,
                "result_dir": _base_dir / RunSubDirType.RESULT,
                "input_config": _base_dir
                / RunSubDirType.CONFIG
                / DEFAULT_INPUT_JSON_FNAME,
            },
        )

    def _do_run_cmd_in_container(self, typ: str) -> None:
        cmd = self._gen_run_container_cmd(typ)
        console.rule(f":elephant: {typ} docker cmd", align="left")
        console.print(f"{cmd}\n")
        console.print(
            f":fish: eval run:{typ} dir @ [green blink]{self._workdir}/{typ}[/]"
        )
        if not self.gencmd:
            check_call(f"docker pull {self.baseimage}", shell=True)
            check_call(cmd, shell=True)

    def _gen_run_container_cmd(self, typ: str) -> str:
        if typ not in (EvalTaskType.PPL, EvalTaskType.CMP):
            raise Exception(f"no support {typ} to gen docker cmd")

        rundir = self._workdir / typ

        cmd = [
            "docker",
            "run",
            "--net=host",
            "--rm",
            "--name",
            f"{self._version}-{typ}",
            "-e",
            "DEBUG=1",
        ]
        cmd += [
            "-v",
            f"{rundir}:{_CNTR_WORKDIR}",
            "-v",
            f"{self._model_dir}:{_CNTR_WORKDIR}/{RunSubDirType.SWMP}/src",
            "-v",
            f"{self._model_dir}/{DefaultYAMLName.MODEL}:{_CNTR_WORKDIR}/{RunSubDirType.SWMP}/{DefaultYAMLName.MODEL}",
            "-v",
            f"{self._runtime_dir}/dep:{_CNTR_WORKDIR}/{RunSubDirType.SWMP}/dep",
            "-v",
            f"{self._runtime_dir}/{DEFAULT_MANIFEST_NAME}:{_CNTR_WORKDIR}/{RunSubDirType.SWMP}/{DEFAULT_MANIFEST_NAME}",
        ]

        if typ == EvalTaskType.PPL:
            cmd += [
                "-v",
                f"{self.project_dir / URIType.DATASET}:{_CNTR_WORKDIR}/{RunSubDirType.DATASET}",
            ]
        elif typ == EvalTaskType.CMP:
            cmd += [
                "-v",
                f"{self._ppl_workdir / RunSubDirType.RESULT}:{_CNTR_WORKDIR}/{RunSubDirType.PPL_RESULT}",
            ]

        cntr_cache_dir = os.environ.get("SW_PIP_CACHE_DIR", CNTR_DEFAULT_PIP_CACHE_DIR)
        host_cache_dir = os.path.expanduser("~/.cache/starwhale-pip")
        cmd += ["-v", f"{host_cache_dir}:{cntr_cache_dir}"]

        _env = os.environ
        for _ee in (
            "SW_PYPI_INDEX_URL",
            "SW_PYPI_EXTRA_INDEX_URL",
            "SW_PYPI_TRUSTED_HOST",
            "SW_RESET_CONDA_CONFIG",
        ):
            if _ee not in _env:
                continue
            cmd.extend(["-e", f"{_ee}={_env[_ee]}"])

        cmd += [self.baseimage, typ]
        return " ".join(cmd)

    def _render_manifest(self) -> None:
        _status = True
        for _d in (self._ppl_workdir, self._cmp_workdir):
            _f = _d / RunSubDirType.STATUS / CURRENT_FNAME
            if not _f.exists():
                continue
            _status = _status and (_f.open().read().strip() == _STATUS.SUCCESS)

        self._manifest.update(
            dict(
                name=self.name,
                desc=self.desc,
                model=self.model_uri,
                model_dir=str(self._model_dir),
                datasets=[u.full_uri for u in self.dataset_uris],
                runtime=self.runtime_uri,
                status=_STATUS.SUCCESS if _status else _STATUS.FAILED,
                finished_at=now_str(),  # type: ignore
            )
        )
        _f = self._workdir / DEFAULT_MANIFEST_NAME
        ensure_file(_f, yaml.safe_dump(self._manifest, default_flow_style=False))



import typing as t

import click

from .view import get_term_view, InstanceTermView


@click.group(
    "instance",
    help="Starwhale Instance management, login and select standalone or cloud instance",
)
@click.pass_context
def instance_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@instance_cmd.command("select")
@click.argument("instance")
def _select(instance: str) -> None:
    """
    Select current default instance

    INSTANCE: use alias name to select
    """
    # TODO: support URI select
    InstanceTermView().select(instance)


@instance_cmd.command("login")
@click.argument("instance", default="")
@click.option("--username", prompt="username", required=True)
@click.password_option(confirmation_prompt=False)
@click.option("--alias", type=str, help="Starwhale instance alias name", required=True)
def _login(instance: str, username: str, password: str, alias: str) -> None:
    """Login Starwhale Instance

    * INSTANCE: Instance URI, if ignore it, swcli will login current selected instance.
    """
    InstanceTermView().login(instance, username, password, alias)


@instance_cmd.command("logout")
@click.argument("instance", default="")
def _logout(instance: str) -> None:
    """
    Logout Starwhale instance

    * INSTANCE: instance alias name or uri, if ignore it, swcli will logout current selected instance.
    """
    click.confirm("Do you want to continue?", abort=True)
    InstanceTermView().logout(instance)


@instance_cmd.command("list", help="List login Starwhale instances")
@click.pass_obj
def _list(view: t.Type[InstanceTermView]) -> None:
    view().list()


@instance_cmd.command("info", help="Show instance details")
@click.argument("instance", default="")
@click.pass_obj
def _info(view: t.Type[InstanceTermView], instance: str) -> None:
    view().info(instance)



from __future__ import annotations

import os
import json
import typing as t
from abc import ABCMeta, abstractmethod
from pathlib import Path

import boto3
import loguru
from loguru import logger as _logger
from botocore.client import Config as S3Config

from starwhale.consts import DataLoaderKind, SWDSBackendType
from starwhale.utils.error import NoSupportError

# TODO: config chunk size
_CHUNK_SIZE = 8 * 1024 * 1024  # 8MB
_FILE_END_POS = -1


class DataField(t.NamedTuple):
    idx: int
    data_size: int
    batch_size: int
    data: bytes
    ext_attr: t.Dict[str, t.Any]


# TODO: use attr to simplify code


class DataLoader(object):

    __metaclass__ = ABCMeta

    def __init__(
        self,
        storage: StorageBackend,
        swds: t.List[t.Dict[str, t.Any]] = [],
        logger: t.Union[loguru.Logger, None] = None,
        kind: str = DataLoaderKind.SWDS,
        deserializer: t.Optional[t.Callable] = None,
    ):
        self.storage = storage
        self.swds = swds
        self.logger = logger or _logger
        self.kind = kind
        self.deserializer = deserializer

        self._do_validate()

    def _do_validate(self) -> None:
        if self.kind not in (DataLoaderKind.JSONL, DataLoaderKind.SWDS):
            raise Exception(f"{self.kind} no support")

    @abstractmethod
    def __iter__(self) -> t.Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return f"DataLoader for {self.storage.backend}"

    def __repr__(self) -> str:
        return (
            f"DataLoader for {self.storage.backend}, "
            f"swds({len(self.swds)}), extra:{self.storage.service}"
        )


class JSONLineDataLoader(DataLoader):
    def __init__(
        self,
        storage: StorageBackend,
        swds: t.List[t.Dict[str, t.Any]] = [],
        logger: t.Union[loguru.Logger, None] = None,
        **kwargs,
    ) -> None:
        super().__init__(storage, swds, logger, DataLoaderKind.JSONL, **kwargs)

    def __iter__(self) -> t.Any:
        for _swds in self.swds:
            for data in self._do_iter(_swds["bucket"], _swds["key"]["data"]):
                yield data

    def _do_iter(self, bucket: str, key_compose: str) -> t.Any:
        self.logger.info(f"@{bucket}/{key_compose}")
        _file = self.storage._make_file(bucket, key_compose)
        while True:
            line = _file.readline()
            if not line:
                break

            line = line.strip()
            if not line:
                continue

            if self.deserializer:
                yield self.deserializer(line)
                continue
            # TODO: add json exception ignore?
            yield json.loads(line)


class SWDSDataLoader(DataLoader):
    def __init__(
        self,
        storage: StorageBackend,
        swds: t.List[t.Dict[str, t.Any]] = [],
        logger: t.Union[loguru.Logger, None] = None,
    ) -> None:
        super().__init__(storage, swds, logger, DataLoaderKind.SWDS)

    def __iter__(self):
        for _swds in self.swds:
            for data, label in zip(
                self._do_iter(
                    _swds["bucket"], _swds["key"]["data"], _swds.get("ext_attr", {})
                ),
                self._do_iter(
                    _swds["bucket"], _swds["key"]["label"], _swds.get("ext_attr", {})
                ),
            ):
                yield data, label

    def _do_iter(
        self, bucket: str, key_compose: str, ext_attr: t.Dict[str, t.Any]
    ) -> t.Iterator[DataField]:
        from .dataset import _header_size, _header_struct

        self.logger.info(f"@{bucket}/{key_compose}")
        _file = self.storage._make_file(bucket, key_compose)
        while True:
            header = _file.read(_header_size)
            if not header:
                break
            _, _, idx, size, padding_size, batch, _ = _header_struct.unpack(header)
            data = _file.read(size + padding_size)
            yield DataField(
                idx,
                size,
                batch,
                data[:size].tobytes() if isinstance(data, memoryview) else data[:size],
                ext_attr,
            )


class StorageBackend(object):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        backend: str,
        secret: t.Dict[str, t.Any] = {},
        service: t.Dict[str, t.Any] = {},
    ) -> None:
        self.service = service
        self.backend = backend
        self.secret = secret

        self._do_validate()

    def __str__(self) -> str:
        return f"StorageBackend for {self.backend}"

    def __repr__(self) -> str:
        return f"StorageBackend for {self.backend}, service: {self.service}"

    def _do_validate(self) -> None:
        # TODO: add more validator
        if self.backend == SWDSBackendType.S3:
            _s = self.secret
            if (
                not _s
                or not isinstance(_s, dict)
                or not _s.get("access_key")
                or not _s.get("secret_key")
            ):
                raise Exception(f"secret({_s}) format is invalid")

            _s = self.service
            if (
                not _s
                or not isinstance(_s, dict)
                or not _s.get("endpoint")
                or not _s.get("region")
            ):
                raise Exception(f"s3_service({_s} format is invalid)")

    def _parse_key(self, key: str) -> t.Tuple[str, int, int]:
        # TODO: some builtin method to
        # TODO: add start end normalize
        _r = key.split(":")
        if len(_r) == 1:
            return _r[0], 0, _FILE_END_POS
        elif len(_r) == 2:
            return _r[0], int(_r[1]), _FILE_END_POS
        else:
            return _r[0], int(_r[1]), int(_r[2])

    # TODO: tune typing hint for FileObj
    @abstractmethod
    def _make_file(self, bucket: str, key_compose: str) -> t.Any:
        raise NotImplementedError


class S3StorageBackend(StorageBackend):
    def __init__(
        self, secret: t.Dict[str, t.Any] = {}, service: t.Dict[str, t.Any] = {}
    ):
        super().__init__(backend=SWDSBackendType.S3, secret=secret, service=service)

        _env = os.environ
        self.s3 = boto3.resource(
            "s3",
            endpoint_url=self.service["endpoint"],
            aws_access_key_id=self.secret["access_key"],
            aws_secret_access_key=self.secret["secret_key"],
            config=S3Config(
                connect_timeout=float(_env.get("SW_S3_CONNECT_TIMEOUT", 10)),
                read_timeout=float(_env.get("SW_S3_READ_TIMEOUT", 60)),
                signature_version="s3v4",
                retries={
                    "total_max_attempts": int(_env.get("SW_S3_TOTAL_MAX_ATTEMPTS", 6)),
                    "mode": "standard",
                },
            ),
            region_name=self.service["region"],
        )

    # TODO: tune return typing hint
    def _make_file(self, bucket: str, key_compose: str) -> t.Any:
        _key, _start, _end = self._parse_key(key_compose)
        return S3BufferedFileLike(
            s3=self.s3,
            bucket=bucket,
            key=_key,
            start=_start,
            end=_end,
        )


class FuseStorageBackend(StorageBackend):
    def __init__(self) -> None:
        super().__init__(backend=SWDSBackendType.FUSE)

    def _make_file(self, bucket: str, key_compose: str) -> t.Any:
        _key, _start, _ = self._parse_key(key_compose)
        bucket_path = (
            Path(bucket).expanduser() if bucket.startswith("~/") else Path(bucket)
        )
        _file = (bucket_path / _key).open("rb")
        _file.seek(_start)
        # TODO: support end
        return _file


# TODO: add mock test
class S3BufferedFileLike(object):
    # TODO: add s3 typing
    def __init__(self, s3: t.Any, bucket: str, key: str, start: int, end: int) -> None:
        self.key = key
        self.obj = s3.Object(bucket, key)
        self.start = start
        self.end = end

        self._buffer = memoryview(bytearray(0))
        self._current = 0
        self._s3_eof = False
        self._current_s3_start = start
        self._iter_lines = None

    def tell(self) -> int:
        return self._current

    def readline(self) -> str:
        if self._iter_lines is None:
            self._iter_lines = self.obj.get()["Body"].iter_lines(chunk_size=_CHUNK_SIZE)

        try:
            line: bytes = next(self._iter_lines)  # type: ignore
        except StopIteration:
            line = b""
        return line.decode()

    def read(self, size: int) -> memoryview:
        # TODO: use smart_open 3rd lib?
        if (self._current + size) <= len(self._buffer):
            end = self._current + size
            out = self._buffer[self._current : end]
            self._current = end
            return out
        else:
            data, _ = self._next_data()
            _release_buffer = self._buffer
            self._buffer = memoryview(self._buffer[self._current :].tobytes() + data)
            _release_buffer.release()
            self._current = 0

            if len(self._buffer) == 0:
                return memoryview(bytearray(0))  # EOF
            elif (self._current + size) > len(self._buffer):
                # TODO: maybe ignore this error?
                raise Exception(
                    f"{self.key} file cannot read {size} data, error format"
                )
            else:
                end = self._current + size
                out = self._buffer[self._current : end]
                self._current = end
                return out

    def close(self) -> None:
        # TODO: cleanup stream and open
        self.obj = None
        try:
            self._buffer.release()
        except Exception as e:
            _logger.warning(f"skip _buffer(memoryview) release exception:{e}")

    def _next_data(self) -> t.Tuple[bytes, int]:
        end = _CHUNK_SIZE + self._current_s3_start - 1
        end = end if self.end == _FILE_END_POS else min(self.end, end)

        data, length = self._do_fetch_data(self._current_s3_start, end)
        self._current_s3_start += length

        return data, length

    def _do_fetch_data(self, _start: int, _end: int) -> t.Tuple[bytes, int]:
        # TODO: add more exception handle
        if self._s3_eof or (_end != _FILE_END_POS and _end < _start):
            return b"", 0

        resp = self.obj.get(Range=f"bytes={_start}-{_end}")
        body = resp["Body"]
        length = resp["ContentLength"]
        out = resp["Body"].read()
        body.close()

        self._s3_eof = _end == _FILE_END_POS or (_end - _start + 1) > length
        return out, length


def get_data_loader(
    swds_config: t.Dict[str, t.Any],
    logger: t.Union[loguru.Logger, None] = None,
    **kwargs,
) -> DataLoader:
    """s3 or fuse data loader

    Args:
        swds_config (dict): origin json example

    {
        "backend": "s3",  // s3 or fuse
        "kind": "swds",       // swds or jsonline
        "secret": {       // auth info
            "access_key": "username or access key",
            "secret_key": "password or secret key"
        },
        "service": {  // only for s3
            "endpoint": "s3 address",
            "region": "s3 region",
        },
        "swds": [   // swds dataset address
            {
                "bucket": "s3 bucket or rootdir",
                "key": {
                    "data":  "{name}:{start_pos}:{end_pos}", // start default is 0
                    "label": "{name}:{start_pos}:{end_pos}"  // end default is -1, which is EOF of File or Object
                },
                "ext_attr":{  //optional
                    "ds_name": "dataset name",
                    "ds_version": "dataset version"
                }
            },
            {
                "bucket": "s3 bucket or rootdir",
                "key": {
                    "data":  "{name}:{start_pos}:{end_pos}",
                    "label": "{name}:{start_pos}:{end_pos}"
                }
            }
        ]
    }

        logger (t.Union[loguru.Logger, None], optional): logger. Defaults to None.

    Raises:
        NoSupportError: _description_

    Returns:
        DataLoader: S3DataLoader or FuseDataLoader
    """
    logger = logger or _logger

    _storage: t.Union[S3StorageBackend, FuseStorageBackend]

    _backend = swds_config["backend"]
    if _backend == SWDSBackendType.S3:
        _storage = S3StorageBackend(swds_config["secret"], swds_config["service"])
    elif _backend == SWDSBackendType.FUSE:
        _storage = FuseStorageBackend()
    else:
        raise NoSupportError(f"{_backend} backend storage, no support")

    _kind = swds_config.get("kind", DataLoaderKind.SWDS)
    if _kind == DataLoaderKind.JSONL:
        return JSONLineDataLoader(_storage, swds_config["swds"], logger, **kwargs)
    elif _kind == DataLoaderKind.SWDS:
        return SWDSDataLoader(_storage, swds_config["swds"], logger)
    else:
        raise NoSupportError(f"{_kind} data loader, no support")



import sys
import webbrowser
from pathlib import Path

import click

from starwhale.utils import console
from starwhale.consts import (
    RECOVER_DIRNAME,
    VERSION_PREFIX_CNT,
    STANDALONE_INSTANCE,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.utils.fs import empty_dir
from starwhale.base.type import URIType, get_bundle_type_by_uri
from starwhale.utils.config import SWCliConfigMixed


def gc(dry_run: bool = False, yes: bool = False) -> None:
    sw = SWCliConfigMixed()

    for project_dir in sw.rootdir.iterdir():
        project_name = project_dir.name
        if project_name == RECOVER_DIRNAME:
            continue

        if not yes and not click.confirm(
            f"continue to garbage collection for project: {project_name}?"
        ):
            continue

        _removed_paths = []
        for typ in (URIType.DATASET, URIType.JOB, URIType.RUNTIME, URIType.MODEL):
            _bundle_type = get_bundle_type_by_uri(typ) if typ != URIType.JOB else ""
            _recover_dir = project_dir / typ / RECOVER_DIRNAME

            if typ in (URIType.RUNTIME, URIType.MODEL):
                for _path in _recover_dir.glob(f"**/*{_bundle_type}"):
                    if not (_path.is_file() and _path.name.endswith(_bundle_type)):
                        continue
                    _removed_paths.append(_path)
                    _workdir_path = _get_workdir_path(project_dir, typ, _path)
                    if _workdir_path.exists():
                        _removed_paths.append(_workdir_path)
            elif typ in (URIType.DATASET, URIType.JOB):
                for _path in _recover_dir.glob(f"**/{DEFAULT_MANIFEST_NAME}"):
                    if not _path.is_file():
                        continue
                    _removed_paths.append(_path.parent)
                    _workdir_path = _get_workdir_path(project_dir, typ, _path.parent)
                    if _workdir_path.exists():
                        _removed_paths.append(_workdir_path)

        _rcnt = len(_removed_paths)
        if _rcnt == 0:
            console.print(f":deer: project:[red]{project_name}[/], no objects to gc")
            continue

        console.print(
            f":zap: project:[red]{project_name}[/], find {len(_removed_paths)} objects to cleanup..."
        )
        console.print("\n".join([f":no_entry_sign: {_p}" for _p in _removed_paths]))

        if not dry_run and (yes or click.confirm("continue to remove?")):
            for _p in _removed_paths:
                empty_dir(_p)

    for project_dir in (sw.rootdir / RECOVER_DIRNAME).iterdir():
        if not dry_run and (yes or click.confirm("continue to remove?")):
            empty_dir(project_dir)

        console.print(f":no_entry_sign: [red]{project_dir}[/] removed")


def _get_workdir_path(project_dir: Path, typ: str, bundle_path: Path) -> Path:
    version = bundle_path.name.split(".")[0]
    object_prefix = f"{version[:VERSION_PREFIX_CNT]}/{version}"

    if typ != URIType.JOB:
        object_name = bundle_path.parent.parent.name
        object_prefix = f"{object_name}/{object_prefix}"
    _rpath = project_dir / "workdir" / typ / RECOVER_DIRNAME / object_prefix
    if _rpath.exists():
        return _rpath
    else:
        return project_dir / "workdir" / typ / object_prefix


def open_web(instance_uri: str = "") -> None:
    sw = SWCliConfigMixed()

    instance_uri = instance_uri.strip().strip("/")
    instance_uri = instance_uri or sw.current_instance

    if instance_uri == STANDALONE_INSTANCE:
        console.print(":see_no_evil: standalone web ui is comming soon.")
    else:
        _config = sw.get_sw_instance_config(instance_uri)
        _uri = _config.get("uri", "")
        if not _uri:
            console.print(
                f":rotating_light: sw cannot find instance [red]{instance_uri}[/], please [bold red]swcli instance login[/] first."
            )
            sys.exit(1)
        else:
            console.print(f":clap: try to open {_uri}")
            webbrowser.open(_uri)



import os
import sys
import shutil
import typing as t
import tarfile
import platform
import subprocess
from pathlib import Path

import conda_pack
import virtualenv
from loguru import logger

from starwhale.utils import console, is_linux, is_darwin, is_windows
from starwhale.consts import (
    ENV_VENV,
    ENV_CONDA,
    PythonRunEnv,
    ENV_CONDA_PREFIX,
    SW_PYPI_PKG_NAME,
    DEFAULT_PYTHON_VERSION,
)
from starwhale.utils.fs import empty_dir, ensure_dir, ensure_file
from starwhale.utils.error import (
    FormatError,
    ExistedError,
    NotFoundError,
    NoSupportError,
    PythonEnvironmentError,
)
from starwhale.utils.process import check_call

CONDA_ENV_TAR = "env.tar.gz"
DUMP_CONDA_ENV_FNAME = "env-lock.yaml"
DUMP_PIP_REQ_FNAME = "requirements-lock.txt"
DUMP_USER_PIP_REQ_FNAME = "requirements.txt"
SUPPORTED_PIP_REQ = [DUMP_USER_PIP_REQ_FNAME, "pip-req.txt", "pip3-req.txt"]

_DUMMY_FIELD = -1


class PythonVersionField(t.NamedTuple):
    major: int = _DUMMY_FIELD
    minor: int = _DUMMY_FIELD
    micro: int = _DUMMY_FIELD


def conda_install_req(
    req: t.Union[str, Path],
    env_name: str = "",
    prefix_path: t.Optional[Path] = None,
    enable_pre: bool = False,
) -> None:
    prefix_cmd = [get_conda_bin(), "run"]

    if env_name:
        prefix_cmd += ["--name", env_name]
    elif prefix_path is not None:
        prefix_cmd += ["--prefix", str(prefix_path.absolute())]

    prefix_cmd += ["python3", "-m", "pip"]
    _do_pip_install_req(prefix_cmd, req, enable_pre)


def _do_pip_install_req(
    prefix_cmd: t.List[t.Any], req: t.Union[str, Path], enable_pre: bool = False
) -> None:
    cmd = prefix_cmd + [
        "install",
        "--exists-action",
        "w",
    ]

    _env = os.environ
    if _env.get("SW_PYPI_INDEX_URL"):
        cmd += [
            "--index-url",
            _env["SW_PYPI_INDEX_URL"],
        ]
    if _env.get("SW_PYPI_EXTRA_INDEX_URL"):
        cmd += [
            "--extra-index-url",
            _env["SW_PYPI_EXTRA_INDEX_URL"],
        ]
    if _env.get("SW_PYPI_TRUSTED_HOST"):
        cmd += ["--trusted-host", _env["SW_PYPI_TRUSTED_HOST"]]

    if enable_pre:
        cmd += ["--pre"]

    if isinstance(req, Path):
        cmd += ["-r", str(req.absolute())]
    elif os.path.isfile(req):
        cmd += ["-r", req]
    else:
        cmd += [req]

    check_call(cmd)


def venv_install_req(
    venvdir: t.Union[str, Path], req: t.Union[str, Path], enable_pre: bool = False
) -> None:
    venvdir = str(venvdir)
    req = str(req)
    prefix_cmd = [os.path.join(venvdir, "bin", "pip")]
    _do_pip_install_req(prefix_cmd, req, enable_pre)


def venv_activate(venvdir: t.Union[str, Path]) -> None:
    _fpath = Path(venvdir) / "bin" / "activate"
    cmd = f"source {_fpath.absolute()}"
    check_call(cmd, shell=True, executable="/bin/bash")


def parse_python_version(s: str) -> PythonVersionField:
    s = s.strip().lower()

    if not s:
        raise FormatError("python version empty")

    if s.startswith("python"):
        s = s.split("python", 1)[-1]

    _vt = s.split(".")
    _major, _minor, _micro = int(_vt[0]), _DUMMY_FIELD, _DUMMY_FIELD

    if len(_vt) >= 2:
        _minor = int(_vt[1])

    if len(_vt) >= 3:
        _micro = int(_vt[2])

    return PythonVersionField(major=_major, minor=_minor, micro=_micro)


def venv_setup(
    venvdir: t.Union[str, Path],
    python_version: str,
    prompt: str = "",
    clear: bool = False,
) -> None:
    # TODO: define starwhale virtualenv.py
    # TODO: use more elegant method to make venv portable
    args = [str(venvdir)]
    if prompt:
        args += ["--prompt", prompt]

    if python_version:
        _v = parse_python_version(python_version)
        args += ["--python", f"{_v.major}.{_v.minor}"]

    if clear:
        args += ["--clear"]

    session = virtualenv.cli_run(args)
    console.print(f":clap: create venv@{venvdir}, python:{session.interpreter.version}")  # type: ignore


def pip_freeze(
    py_env: str, path: t.Union[str, Path], include_editable: bool = False
) -> None:
    # TODO: add cmd timeout and error log
    _py_bin = get_user_runtime_python_bin(py_env)
    logger.info(f"{_py_bin}: pip freeze")
    cmd = [_py_bin, "-m", "pip", "freeze", "--require-virtualenv"]
    if not include_editable:
        cmd += ["--exclude-editable"]
    cmd += [">", str(path)]

    check_call(" ".join(cmd), shell=True)


def user_pip_install_pkg(py_env: str, pkg_name: str, enable_pre: bool = False) -> None:
    _py_bin = get_user_runtime_python_bin(py_env)
    cmd = [_py_bin, "-m", "pip", "install"]

    if enable_pre:
        cmd += ["--pre"]

    cmd += [pkg_name]
    check_call(cmd)


def check_python_interpreter_consistency(mode: str) -> t.Tuple[bool, str, str]:
    if mode == PythonRunEnv.CONDA:
        ep_base_prefix = os.environ.get(ENV_CONDA_PREFIX, "")
    elif mode == PythonRunEnv.VENV:
        ep_base_prefix = os.environ.get(ENV_VENV, "")
    else:
        ep_base_prefix = (
            os.environ.get(ENV_VENV)
            or os.environ.get(ENV_CONDA_PREFIX)
            or sys.base_prefix
        )
    logger.debug(
        f"current python interpreter base_prefix:{sys.base_prefix}, expected env base_prefix:{ep_base_prefix}"
    )
    _ok = ep_base_prefix == sys.base_prefix
    if not _ok:
        cur_version = f"{sys.version_info.major}.{sys.version_info.minor}"
        user_version = get_user_python_version(mode)
        if not user_version.startswith(cur_version):
            logger.error(
                f"swcli use python:{cur_version}, but runtime venv/conda python:{user_version}"
            )
            raise PythonEnvironmentError(
                f"swcli({cur_version}), runtime({user_version})"
            )

    return _ok, sys.base_prefix, ep_base_prefix


def guess_current_py_env() -> str:
    if is_venv():
        return PythonRunEnv.VENV
    elif is_conda():
        return PythonRunEnv.CONDA
    else:
        return PythonRunEnv.SYSTEM


def get_user_python_sys_paths(py_env: str) -> t.List[str]:
    logger.debug(f"get env({py_env}) sys path")
    _py_bin = get_user_runtime_python_bin(py_env)
    logger.info(f"{_py_bin}: sys.path")
    output = subprocess.check_output(
        [
            _py_bin,
            "-c",
            "import sys; print(','.join(sys.path))",
        ]
    )
    return output.decode().strip().split(",")


def conda_create(
    env: str,
    python_version: str = DEFAULT_PYTHON_VERSION,
    quiet: bool = False,
) -> None:
    cmd = ["conda", "create", "--name", env, "--yes"]

    if quiet:
        cmd += ["--quiet"]

    cmd += [f"python={python_version}"]
    check_call(cmd)


def conda_export(path: t.Union[str, Path], env: str = "") -> None:
    # TODO: add cmd timeout
    cmd = f"{get_conda_bin()} env export"
    env = f"-n {env}" if env else ""
    check_call(f"{cmd} {env} > {path}", shell=True)


def conda_restore(
    env_fpath: t.Union[str, Path], target_env: t.Union[str, Path]
) -> None:
    cmd = f"{get_conda_bin()} env update --file {env_fpath} --prefix {target_env}"
    check_call(cmd, shell=True)


def conda_activate(env: t.Union[str, Path]) -> None:
    cmd = f"{get_conda_bin()} activate {env}"
    check_call(cmd, shell=True)


def conda_activate_render(env_dir: Path, workdir: Path) -> None:
    sw_cntr_content = """
_conda_hook="$(/opt/miniconda3/bin/conda shell.bash hook)"
cat >> /dev/stdout << EOF
$_conda_hook
conda activate /opt/starwhale/swmp/dep/conda/env
EOF
"""

    host_content = f"""
echo 'conda activate {env_dir.absolute()}'
"""
    _render_sw_activate(sw_cntr_content, host_content, workdir)


def venv_activate_render(
    venvdir: t.Union[str, Path], workdir: Path, relocate: bool = False
) -> None:
    bin = f"{venvdir}/bin"
    host_content = f"""
echo 'source {venvdir}/bin/activate'
"""

    if relocate:
        # TODO: support relocatable editable python package
        sw_cntr_content = f"""
sed -i '1d' {bin}/starwhale {bin}/sw {bin}/swcli {bin}/pip* {bin}/virtualenv
sed -i '1i\#!{bin}/python3' {bin}/starwhale {bin}/sw {bin}/swcli {bin}/pip* {bin}/virtualenv

sed -i 's#^VIRTUAL_ENV=.*$#VIRTUAL_ENV={venvdir}#g' {bin}/activate
rm -rf {bin}/python3
ln -s /usr/bin/python3 {bin}/python3
echo 'source {bin}/activate'
"""
    else:
        sw_cntr_content = host_content

    _render_sw_activate(sw_cntr_content, host_content, workdir)


def _render_sw_activate(sw_cntr_content: str, host_content: str, workdir: Path) -> None:
    _sw_path = workdir / "activate.sw"
    _host_path = workdir / "activate.host"

    ensure_file(_sw_path, sw_cntr_content, mode=0o755)
    ensure_file(_host_path, host_content, mode=0o755)

    console.print(
        f" :clap: {_sw_path.name} and {_host_path.name} is generated at {workdir}"
    )
    console.print(" :compass: run cmd:  ")
    console.print(f" \t Docker Container: [bold red] $(sh {_sw_path}) [/]")
    console.print(f" \t Host: [bold red] $(sh {_host_path}) [/]")


def get_conda_bin() -> str:
    # TODO: add process cache
    for _p in (
        "/opt/miniconda3/bin/conda",
        "/opt/anaconda3/bin/conda",
        os.path.expanduser("~/miniconda3/bin/conda"),
        os.path.expanduser("~/anaconda3/bin/conda"),
    ):
        if os.path.exists(_p):
            return _p
    else:
        return "conda"


def dump_python_dep_env(
    dep_dir: t.Union[str, Path],
    pip_req_fpath: str,
    gen_all_bundles: bool = False,
    expected_runtime: str = "",
    mode: str = PythonRunEnv.AUTO,
    include_editable: bool = False,
    identity: str = "",
) -> t.Dict[str, t.Any]:
    # TODO: smart dump python dep by starwhale sdk-api, pip ast analysis?
    dep_dir = Path(dep_dir)

    pr_env = get_python_run_env(mode)
    sys_name = platform.system()
    py_ver = get_user_python_version(pr_env)

    validate_python_environment(mode, expected_runtime, identity)
    validate_runtime_package_dep(mode)

    _manifest = dict(
        expected_mode=mode,
        env=pr_env,
        system=sys_name,
        python=py_ver,
        local_gen_env=False,
        venv=dict(use=is_venv()),
        conda=dict(use=is_conda()),
    )

    _conda_dir = dep_dir / "conda"
    _python_dir = dep_dir / "python"
    _venv_dir = _python_dir / "venv"
    _pip_lock_req = _python_dir / DUMP_PIP_REQ_FNAME
    _conda_lock_env = _conda_dir / DUMP_CONDA_ENV_FNAME

    ensure_dir(_venv_dir)
    ensure_dir(_conda_dir)
    ensure_dir(_python_dir)

    console.print(
        f":dizzy: python{py_ver}@{pr_env}, os({sys_name}), include-editable({include_editable}), try to export environment..."
    )

    if os.path.exists(pip_req_fpath):
        shutil.copyfile(pip_req_fpath, str(_python_dir / DUMP_USER_PIP_REQ_FNAME))

    if pr_env == PythonRunEnv.CONDA:
        if include_editable:
            raise NoSupportError("conda cannot support export pip editable package")

        logger.info(f"[info:dep]dump conda environment yaml: {_conda_lock_env}")
        conda_export(_conda_lock_env)
    elif pr_env == PythonRunEnv.VENV:
        logger.info(f"[info:dep]dump pip-req with freeze: {_pip_lock_req}")
        pip_freeze(pr_env, _pip_lock_req, include_editable)
    else:
        # TODO: add other env tools
        logger.warning(
            "detect use system python, swcli does not pip freeze, only use custom pip-req"
        )

    if is_windows() or is_darwin() or not gen_all_bundles:
        # TODO: win/osx will produce env in controller agent with task
        logger.info(f"[info:dep]{sys_name} will skip conda/venv dump or generate")
    elif is_linux():
        # TODO: more design local or remote build venv
        # TODO: ignore some pkg when dump, like notebook?
        _manifest["local_gen_env"] = True  # type: ignore

        if pr_env == PythonRunEnv.CONDA:
            cenv = get_conda_env()
            dest = str(_conda_dir / CONDA_ENV_TAR)
            if not cenv:
                raise Exception("cannot get conda env value")

            # TODO: add env/env-name into model.yaml, user can set custom vars.
            logger.info("[info:dep]try to pack conda...")
            conda_pack.pack(
                name=cenv,
                force=True,
                output=dest,
                ignore_editable_packages=not include_editable,
            )
            logger.info(f"[info:dep]finish conda pack {dest})")
            console.print(f":beer_mug: conda pack @ [underline]{dest}[/]")
        else:
            # TODO: tune venv create performance, use clone?
            logger.info(f"[info:dep]build venv dir: {_venv_dir}")
            venv_setup(_venv_dir, python_version=expected_runtime)
            logger.info(
                f"[info:dep]install pip freeze({_pip_lock_req}) to venv: {_venv_dir}"
            )
            venv_install_req(_venv_dir, _pip_lock_req)
            if os.path.exists(pip_req_fpath):
                logger.info(
                    f"[info:dep]install custom pip({pip_req_fpath}) to venv: {_venv_dir}"
                )
                # TODO: support ignore editable package
                venv_install_req(_venv_dir, pip_req_fpath)
            console.print(f":beer_mug: venv @ [underline]{_venv_dir}[/]")

    else:
        raise NoSupportError(f"no support {sys_name} system")

    return _manifest


def detect_pip_req(workdir: t.Union[str, Path], fname: str = "") -> str:
    workdir = Path(workdir)

    if fname and (workdir / fname).exists():
        return str(workdir / fname)
    else:
        for p in SUPPORTED_PIP_REQ:
            if (workdir / p).exists():
                return str(workdir / p)
            else:
                return ""
    return ""


def activate_python_env(
    mode: str,
    identity: str,
) -> None:
    # TODO: switch shell python environment directly
    console.print(":cake: run command in shell :cake:")

    if mode == PythonRunEnv.VENV:
        cmd = f"source {identity}/bin/activate"
    elif mode == PythonRunEnv.CONDA:
        cmd = f"conda activate {identity}"
    else:
        raise NoSupportError(mode)

    console.print(f"\t[red][blod]{cmd}")


def create_python_env(
    mode: str,
    name: str,
    workdir: Path,
    python_version: str = DEFAULT_PYTHON_VERSION,
    force: bool = False,
) -> str:

    if mode == PythonRunEnv.VENV:
        venvdir = workdir / "venv"
        if venvdir.exists() and not force:
            raise ExistedError(str(venvdir))

        logger.info(f"create venv @ {venvdir}...")
        venv_setup(venvdir, python_version=python_version, prompt=name)
        return str(venvdir.absolute())
    elif mode == PythonRunEnv.CONDA:
        logger.info(f"create conda {name}:{workdir}, use python {python_version}...")
        conda_create(name, python_version)
        return name
    else:
        raise NoSupportError(mode)


def get_user_python_version(py_env: str) -> str:
    _py_bin = get_user_runtime_python_bin(py_env)
    logger.info(f"{_py_bin}: python version")
    output = subprocess.check_output(
        [
            _py_bin,
            "-c",
            "import sys; _v=sys.version_info;print(f'{_v.major}.{_v.minor}.{_v.micro}')",
        ]
    )
    return output.decode().strip()


def get_user_runtime_python_bin(py_env: str) -> str:
    _prefix = get_base_prefix(py_env)
    _py_bin = os.path.join(_prefix, "bin", "python3")
    if not os.path.exists(_py_bin):
        raise NotFoundError(_py_bin)

    return _py_bin


def get_base_prefix(py_env: str) -> str:
    if py_env == PythonRunEnv.VENV:
        _path = os.environ.get(ENV_VENV, "")
    elif py_env == PythonRunEnv.CONDA:
        _path = os.environ.get(ENV_CONDA_PREFIX, "")
    else:
        _path = sys.prefix

    if _path and os.path.exists(_path):
        return _path
    else:
        raise NotFoundError(f"mode:{py_env}, base_prefix:{_path}")


def is_venv() -> bool:
    # TODO: refactor for get external venv attr
    output = subprocess.check_output(
        [
            "python3",
            "-c",
            "import sys; print(sys.prefix != (getattr(sys, 'base_prefix', None) or (getattr(sys, 'real_prefix', None) or sys.prefix)))",  # noqa: E501
        ],
    )
    return "True" in output.decode() or get_venv_env() != ""


def get_venv_env() -> str:
    return os.environ.get(ENV_VENV, "")


def is_conda() -> bool:
    return get_conda_env() != "" and get_conda_env_prefix() != ""


def get_python_run_env(mode: str = PythonRunEnv.AUTO) -> str:
    if mode == PythonRunEnv.VENV:
        if is_venv():
            return PythonRunEnv.VENV
        else:
            raise EnvironmentError(
                "expected venv mode, but cannot find venv environment"
            )
    elif mode == PythonRunEnv.CONDA:
        if is_conda():
            return PythonRunEnv.CONDA
        else:
            raise EnvironmentError("expected conda mode, but cannot find conda envs")
    elif mode == PythonRunEnv.AUTO:
        if is_conda() and is_venv():
            raise EnvironmentError("find venv and conda both activate")

        if is_conda():
            return PythonRunEnv.CONDA
        elif is_venv():
            return PythonRunEnv.VENV
        else:
            return PythonRunEnv.SYSTEM
    else:
        raise NoSupportError(f"python run env: {mode}")


def get_conda_env() -> str:
    return os.environ.get(ENV_CONDA, "")


def get_conda_env_prefix() -> str:
    return os.environ.get(ENV_CONDA_PREFIX, "")


def restore_python_env(
    workdir: Path,
    mode: str,
    python_version: str,
    local_gen_env: bool = False,
    pip_req: str = "",
) -> None:
    console.print(
        f":bread: restore python:{python_version} {mode}@{workdir}, use local env data: {local_gen_env}"
    )
    _f = _do_restore_conda if mode == PythonRunEnv.CONDA else _do_restore_venv
    _f(workdir, local_gen_env, python_version, pip_req)


def _do_restore_conda(
    _workdir: Path,
    _local_gen_env: bool,
    _python_version: str,
    _pip_req: str,
) -> None:
    _conda_dir = _workdir / "dep" / "conda"
    _tar_fpath = _conda_dir / CONDA_ENV_TAR
    _env_dir = _conda_dir / "env"
    _python_dir = _workdir / "dep" / "python"

    if _local_gen_env and _tar_fpath.exists():
        empty_dir(_env_dir)
        ensure_dir(_env_dir)
        logger.info(f"extract {_tar_fpath} ...")
        with tarfile.open(str(_tar_fpath)) as f:
            f.extractall(str(_env_dir))
        venv_activate_render(_env_dir, _workdir)
    else:
        logger.info("restore conda env ...")
        _env_yaml = _conda_dir / DUMP_CONDA_ENV_FNAME
        conda_restore(_env_yaml, _env_dir)
        _pip_req_path = _python_dir / _pip_req
        if _pip_req_path.exists():
            conda_install_req(req=_pip_req_path, prefix_path=_env_dir)
        conda_activate_render(_env_dir, _workdir)


def _do_restore_venv(
    _workdir: Path,
    _local_gen_env: bool,
    _python_version: str,
    pip_req: str,
    _rebuild: bool = False,
) -> None:
    _python_dir = _workdir / "dep" / "python"
    _venv_dir = _python_dir / "venv"

    _relocate = True
    if _rebuild or not _local_gen_env or not (_venv_dir / "bin" / "activate").exists():
        logger.info(f"setup venv and pip install {_venv_dir}")
        _relocate = False
        venv_setup(_venv_dir, python_version=_python_version)
        for _name in (DUMP_PIP_REQ_FNAME, pip_req):
            _path = _python_dir / _name
            if not _name or not _path.exists():
                continue

            logger.info(f"pip install {_path} ...")
            venv_install_req(_venv_dir, _path)

    venv_activate_render(_venv_dir, _workdir, relocate=_relocate)


def validate_runtime_package_dep(py_env: str) -> None:
    _py_bin = get_user_runtime_python_bin(py_env)
    logger.info(f"{_py_bin}: check {SW_PYPI_PKG_NAME} install")
    cmd = [
        _py_bin,
        "-c",
        f"import pkg_resources; pkg_resources.get_distribution('{SW_PYPI_PKG_NAME}')",
    ]
    try:
        check_call(cmd)
    except subprocess.CalledProcessError:
        console.print(
            f":confused_face: Please install {SW_PYPI_PKG_NAME} in {py_env}, cmd:"
        )
        console.print(
            f"\t :cookie: python3 -m pip install --pre {SW_PYPI_PKG_NAME} :cookie:"
        )
        raise


def validate_python_environment(mode: str, py_version: str, identity: str = "") -> None:
    # TODO: add os platform check
    current_py_env = get_python_run_env(mode)
    current_py_version = get_user_python_version(current_py_env)

    if py_version and not current_py_version.startswith(py_version):
        raise EnvironmentError(
            f"expected python({py_version}) is not equal to detected python({current_py_version})"
        )

    if current_py_env != mode:
        raise EnvironmentError(
            f"expected mode({mode}), detected mode({current_py_env})"
        )

    # TODO: add venv identity check
    if mode == PythonRunEnv.CONDA and not identity:
        conda_name = os.environ.get(ENV_CONDA, "")
        if conda_name != identity:
            raise EnvironmentError(
                f"expected conda name({identity}), detected current conda name({conda_name})"
            )



import typing as t

import click

from starwhale.consts import (
    PythonRunEnv,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    DEFAULT_PYTHON_VERSION,
)

from .view import get_term_view, RuntimeTermView


@click.group(
    "runtime", help="Runtime management, create/build/copy/activate/restore..."
)
@click.pass_context
def runtime_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@runtime_cmd.command(
    "create",
    help="[ONLY Standalone]Create a python runtime, which help user create a easy-to-use and unambiguous environment with venv or conda",
)
@click.argument("workdir")
@click.option("-n", "--name", required=True, help="Runtime name")
@click.option(
    "-m",
    "--mode",
    type=click.Choice([PythonRunEnv.CONDA, PythonRunEnv.VENV]),
    default=PythonRunEnv.VENV,
    help="Runtime mode",
)
@click.option("--python", default=DEFAULT_PYTHON_VERSION, help="Python Version")
@click.option("-f", "--force", is_flag=True, help="Force to create runtime")
def _create(workdir: str, name: str, mode: str, python: str, force: bool) -> None:
    RuntimeTermView.create(
        workdir=workdir,
        name=name,
        mode=mode,
        python_version=python,
        force=force,
    )


@runtime_cmd.command(
    "build",
    help="[ONLY Standalone]Create and build a relocated, shareable, packaged runtime bundle. Support python and native libs.",
)
@click.argument("workdir", type=click.Path(exists=True, file_okay=False))
@click.option("-p", "--project", default="", help="Project URI")
@click.option(
    "-f",
    "--runtime-yaml",
    default=DefaultYAMLName.RUNTIME,
    help="Runtime yaml filename, default use ${workdir}/runtime.yaml file",
)
@click.option(
    "--gen-all-bundles", is_flag=True, help="Generate conda or venv files into runtime"
)
@click.option("--include-editable", is_flag=True, help="Include editable packages")
def _build(
    workdir: str,
    project: str,
    runtime_yaml: str,
    gen_all_bundles: bool,
    include_editable: bool,
) -> None:
    RuntimeTermView.build(
        workdir=workdir,
        project=project,
        yaml_name=runtime_yaml,
        gen_all_bundles=gen_all_bundles,
        include_editable=include_editable,
    )


@runtime_cmd.command("remove")
@click.argument("runtime")
@click.option("-f", "--force", is_flag=True, help="Force to remove runtime")
def _remove(runtime: str, force: bool) -> None:
    """
    Remove runtime

    You can run `swcli runtime recover` to recover the removed runtimes.

    RUNTIME: argument use the `Runtime URI` format, so you can remove the whole runtime or a specified-version runtime.
    """
    click.confirm("continue to remove?", abort=True)
    RuntimeTermView(runtime).remove(force)


@runtime_cmd.command("recover")
@click.argument("runtime")
@click.option("-f", "--force", is_flag=True, help="Force to recover runtime")
def _recover(runtime: str, force: bool) -> None:
    """
    Recover runtime

    RUNTIME: argument use the `Runtime URI` format, so you can recover the whole runtime or a specified-version runtime.
    """
    RuntimeTermView(runtime).recover(force)


@runtime_cmd.command("info", help="Show runtime details")
@click.argument("runtime")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _info(view: t.Type[RuntimeTermView], runtime: str, fullname: bool) -> None:
    view(runtime).info(fullname)


@runtime_cmd.command("history", help="Show runtime history")
@click.argument("runtime", required=True)
@click.option("--fullname", is_flag=True, help="Show version fullname")
def _history(runtime: str, fullname: bool) -> None:
    RuntimeTermView(runtime).history(fullname)


@runtime_cmd.command("restore")
@click.argument("target")
def _restore(target: str) -> None:
    """
    [ONLY Standalone]Prepare dirs, restore python environment with virtualenv or conda and show activate command.

    TARGET: runtime uri or runtime workdir path, in Starwhale Agent Docker Environment, only support workdir path.
    """
    RuntimeTermView.restore(target)


@runtime_cmd.command("list", help="List runtime")
@click.option("--project", default="", help="Project URI")
@click.option("--fullname", is_flag=True, help="Show fullname of runtime version")
@click.option("--show-removed", is_flag=True, help="Show removed runtime")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for tasks list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for tasks list"
)
@click.pass_obj
def _list(
    view: t.Type[RuntimeTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
) -> None:
    view.list(project, fullname, show_removed, page, size)


@runtime_cmd.command(
    "extract", help="[ONLY Standalone]Extract local runtime tar file into workdir"
)
@click.argument("runtime")
@click.option("-f", "--force", is_flag=True, help="Force to extract runtime")
@click.option(
    "--target-dir",
    default="",
    help="Extract target dir.if omitted, sw will use starwhale default workdir",
)
def _extract(runtime: str, force: bool, target_dir: str) -> None:
    RuntimeTermView(runtime).extract(force, target_dir)


@runtime_cmd.command("copy", help="Copy runtime, standalone <--> cloud")
@click.argument("src")
@click.argument("dest")
@click.option("-f", "--force", is_flag=True, help="Force to copy")
def _copy(src: str, dest: str, force: bool) -> None:
    RuntimeTermView.copy(src, dest, force)


@runtime_cmd.command("tag", help="Runtime Tag Management, add or remove")
@click.argument("runtime")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
def _tag(runtime: str, tags: t.List[str], remove: bool, quiet: bool) -> None:
    RuntimeTermView(runtime).tag(tags, remove, quiet)


@runtime_cmd.command(
    "activate",
    help="[Only Standalone]Activate python runtime environment for development",
)
@click.option(
    "-f",
    "--runtime-yaml",
    default=DefaultYAMLName.RUNTIME,
    help="Runtime yaml filename, default use ${WORKDIR}/runtime.yaml file",
)
@click.argument("workdir")
def _activate(workdir: str, runtime_yaml: str) -> None:
    RuntimeTermView.activate(workdir, runtime_yaml)



import pathlib

# TODO: use str path, not Path Class
HOMEDIR = pathlib.Path.home()
CONFIG_DIR = HOMEDIR / ".config" / "starwhale"
SW_CLI_CONFIG = CONFIG_DIR / "config.yaml"

ENV_SW_CLI_CONFIG = "SW_CLI_CONFIG"
ENV_LOG_LEVEL = "SW_LOG_LEVEL"
ENV_SW_IMAGE_REPO = "SW_IMAGE_REPO"

DEFAULT_STARWHALE_API_VERSION = "1.0"
DEFAULT_MANIFEST_NAME = "_manifest.yaml"
DEFAULT_LOCAL_SW_CONTROLLER_ADDR = "localhost:7827"
LOCAL_FUSE_JSON_NAME = "local_fuse.json"
DEFAULT_INPUT_JSON_FNAME = "input.json"
LOCAL_CONFIG_VERSION = "2.0"

# TODO: use ~/.starwhale or ~/.cache/starwhale?
SW_LOCAL_STORAGE = HOMEDIR / ".cache/starwhale"

ENV_CONDA = "CONDA_DEFAULT_ENV"
ENV_CONDA_PREFIX = "CONDA_PREFIX"
ENV_VENV = "VIRTUAL_ENV"


class DefaultYAMLName:
    MODEL = "model.yaml"
    DATASET = "dataset.yaml"
    RUNTIME = "runtime.yaml"


class PythonRunEnv:
    CONDA = "conda"
    VENV = "venv"
    SYSTEM = "system"
    DOCKER = "docker"
    AUTO = "auto"


class HTTPMethod:
    GET = "GET"
    OPTIONS = "OPTIONS"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


class UserRoleType:
    ADMIN = "admin"
    NORMAL = "normal"


FMT_DATETIME = "%Y-%m-%d %H:%M:%S %Z"

# TODO: use better DEFAULT words?
DEFAULT_COPY_WORKERS = 4

JSON_INDENT = 4

SW_API_VERSION = "v1"

SHORT_VERSION_CNT = 12
VERSION_PREFIX_CNT = 2


class SWDSBackendType:
    S3 = "s3"
    FUSE = "fuse"


class DataLoaderKind:
    SWDS = "swds"
    JSONL = "jsonl"


class SWDSSubFileType:
    BIN = "swds_bin"
    META = "swds_meta"


SWDS_DATA_FNAME_FMT = "data_ubyte_{index}.%s" % SWDSSubFileType.BIN
SWDS_LABEL_FNAME_FMT = "label_ubyte_{index}.%s" % SWDSSubFileType.BIN

CURRENT_FNAME = "current"

STANDALONE_INSTANCE = "local"
DEFAULT_INSTANCE = STANDALONE_INSTANCE
DEFAULT_PROJECT = "self"

DEFAULT_PAGE_IDX = 1
DEFAULT_PAGE_SIZE = 20

RECOVER_DIRNAME = ".recover"

DEFAULT_PYTHON_VERSION = "3.8"

YAML_TYPES = (".yaml", ".yml")

DEFAULT_IMAGE_REPO = "ghcr.io/star-whale"
SW_IMAGE_FMT = "{repo}/starwhale:{tag}"
DEFAULT_SW_TASK_RUN_IMAGE = SW_IMAGE_FMT.format(repo=DEFAULT_IMAGE_REPO, tag="latest")
SW_IGNORE_FILE_NAME = ".swignore"

CNTR_DEFAULT_PIP_CACHE_DIR = "/root/.cache/pip"

SW_DEV_DUMMY_VERSION = "0.0.0.dev0"
SW_PYPI_PKG_NAME = "starwhale"



import typing as t

import click

from starwhale.consts import (
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    LOCAL_FUSE_JSON_NAME,
)

from .view import get_term_view, DatasetTermView


@click.group("dataset", help="Dataset management, build/info/list/copy/tag...")
@click.pass_context
def dataset_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@dataset_cmd.command("build", help="[Only Standalone]Build swds with dataset.yaml")
@click.argument("workdir", type=click.Path(exists=True, file_okay=False))
@click.option("-p", "--project", default="", help="Project URI")
@click.option(
    "-f",
    "--dataset-yaml",
    default=DefaultYAMLName.DATASET,
    help="Dataset yaml filename, default use ${WORKDIR}/dataset.yaml file",
)
@click.pass_obj
def _build(
    view: DatasetTermView, workdir: str, project: str, dataset_yaml: str
) -> None:
    # TODO: add cmd options for dataset build, another choice for dataset.yaml
    # TODO: add dry-run
    # TODO: add compress args
    view.build(workdir, project, dataset_yaml)


@dataset_cmd.command("list", help="List dataset")
@click.option("-p", "--project", default="", help="Project URI")
@click.option("--fullname", is_flag=True, help="Show fullname of dataset version")
@click.option("--show-removed", is_flag=True, help="Show removed datasets")
@click.option("--fullname", is_flag=True, help="show version fullname")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for dataset list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for dataset list"
)
@click.pass_obj
def _list(
    view: DatasetTermView,
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
) -> None:
    view.list(project, fullname, show_removed, page, size)


@dataset_cmd.command("info", help="Show dataset details")
@click.argument("dataset")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _info(view: t.Type[DatasetTermView], dataset: str, fullname: bool) -> None:
    view(dataset).info(fullname)


@dataset_cmd.command("remove")
@click.argument("dataset")
@click.option("-f", "--force", is_flag=True, help="Force to remove dataset")
@click.pass_obj
def _remove(view: t.Type[DatasetTermView], dataset: str, force: bool) -> None:
    """
    Remove dataset

    You can run `swcli dataset recover` to recover the removed datasets.

    DATASET: argument use the `Dataset URI` format, so you can remove the whole dataset or a specified-version dataset.
    """
    click.confirm("continue to remove?", abort=True)
    view(dataset).remove(force)


@dataset_cmd.command("recover")
@click.argument("dataset")
@click.option("-f", "--force", is_flag=True, help="Force to recover dataset")
@click.pass_obj
def _recover(view: t.Type[DatasetTermView], dataset: str, force: bool) -> None:
    """
    Recover dataset

    DATASET: argument use the `Dataset URI` format, so you can recover the whole dataset or a specified-version dataset.
    """
    view(dataset).recover(force)


@dataset_cmd.command("history", help="Show dataset history")
@click.argument("dataset")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _history(
    view: t.Type[DatasetTermView], dataset: str, fullname: bool = False
) -> None:
    view(dataset).history(fullname)


@dataset_cmd.command("render-fuse")
@click.argument("target")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help=f"Force to render, if {LOCAL_FUSE_JSON_NAME} was already existed",
)
@click.pass_obj
def _render_fuse(view: t.Type[DatasetTermView], target: str, force: bool) -> None:
    """
    [ONLY Standalone]Render Dataset fuse input.json for standalone ppl

    TARGET: dataset uri or dataset workdir path
    """

    view.render_fuse_json(target, force)


@dataset_cmd.command("copy", help="Copy dataset, standalone <--> cloud")
@click.argument("src")
@click.argument("dest")
@click.option("-f", "--force", is_flag=True, help="Force copy dataset")
@click.pass_obj
def _copy(view: t.Type[DatasetTermView], src: str, dest: str, force: bool) -> None:
    view.copy(src, dest, force)


@dataset_cmd.command("tag", help="Dataset tag management, add or remove")
@click.argument("dataset")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
@click.pass_obj
def _tag(
    view: t.Type[DatasetTermView],
    dataset: str,
    tags: t.List[str],
    remove: bool,
    quiet: bool,
) -> None:
    view(dataset).tag(tags, remove, quiet)



import typing as t
from functools import wraps


class Dict2Obj(object):
    def __init__(self, d: dict) -> None:
        for (k, v) in d.items():
            # TODO: use iterable type
            if isinstance(v, (list, tuple, set)):
                setattr(self, k, [Dict2Obj(i) if isinstance(i, dict) else i for i in v])
            else:
                setattr(self, k, Dict2Obj(v) if isinstance(v, dict) else v)


class Obj2Dict(object):
    def __call__(self, func: t.Any) -> t.Any:
        @wraps(func)
        def _wraps(*args, **kwargs):
            content = func(*args, **kwargs)
            if isinstance(content, (list, tuple, set)):
                return [m if isinstance(m, dict) else m.as_dict() for m in content]
            elif isinstance(content, dict):
                return content
            else:
                # TODO: replace as_dict
                return content.as_dict()

        return _wraps



import math
import struct
import typing as t
from abc import ABCMeta, abstractmethod
from types import TracebackType
from pathlib import Path
from binascii import crc32

import jsonlines

from starwhale.consts import SWDS_DATA_FNAME_FMT, SWDS_LABEL_FNAME_FMT
from starwhale.core.dataset.dataset import (
    D_ALIGNMENT_SIZE,
    D_USER_BATCH_SIZE,
    D_FILE_VOLUME_SIZE,
)

# TODO: tune header size
_header_magic = struct.unpack(">I", b"SWDS")[0]
_data_magic = struct.unpack(">I", b"SDWS")[0]
_header_struct = struct.Struct(">IIQIIII")
_header_size = _header_struct.size


class BuildExecutor(object):
    """
    BuildExecutor can build swds.

    swds_bin format:
        header_magic  uint32  I
        crc           uint32  I
        idx           uint64  Q
        size          uint32  I
        padding_size  uint32  I
        batch_size    uint32  I
        data_magic    uint32  I --> above 32 bytes
        data bytes...
        padding bytes...        --> default 4K padding
    """

    # TODO: add more docstring for class

    __metaclass__ = ABCMeta

    INDEX_NAME = "index.jsonl"
    _DATA_TMP_IDX = "_tmp_index_data.jsonl"
    _LABEL_TMP_IDX = "_tmp_index_label.jsonl"
    _DATA_FMT = SWDS_DATA_FNAME_FMT
    _LABEL_FMT = SWDS_LABEL_FNAME_FMT

    def __init__(
        self,
        data_dir: Path = Path("."),
        output_dir: Path = Path("./sw_output"),
        data_filter: str = "*",
        label_filter: str = "*",
        batch: int = D_USER_BATCH_SIZE,
        alignment_bytes_size: int = D_ALIGNMENT_SIZE,
        volume_bytes_size: int = D_FILE_VOLUME_SIZE,
    ) -> None:
        # TODO: add more docstring for args
        # TODO: validate group upper and lower?
        self._batch = max(batch, 1)
        self.data_dir = data_dir
        self.data_filter = data_filter
        self.label_filter = label_filter
        self.output_dir = output_dir
        self.alignment_bytes_size = alignment_bytes_size
        self.volume_bytes_size = volume_bytes_size

        self._index_writer: t.Optional[jsonlines.Writer] = None
        self._prepare()

    def _prepare(self) -> None:
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self._index_writer = jsonlines.open(
            str((self.output_dir / self.INDEX_NAME).resolve()), mode="w"
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:
            print(f"type:{type}, exception:{value}, traceback:{trace}")

        try:
            self._index_writer.close()  # type: ignore
        except Exception as e:
            print(f"index writer close exception: {e}")

        print("cleanup done.")

    def _write(self, writer: t.Any, idx: int, data: bytes) -> t.Tuple[int, int]:
        size = len(data)
        crc = crc32(data)  # TODO: crc is right?
        start = writer.tell()
        padding_size = self._get_padding_size(size + _header_size)

        _header = _header_struct.pack(
            _header_magic, crc, idx, size, padding_size, self._batch, _data_magic
        )
        _padding = b"\0" * padding_size
        writer.write(_header + data + _padding)
        return start, _header_size + size + padding_size

    def _get_padding_size(self, size: int) -> int:
        remain = (size + _header_size) % self.alignment_bytes_size
        return 0 if remain == 0 else (self.alignment_bytes_size - remain)

    def _write_index(
        self,
        idx: int,
        fno: int,
        data_pos: int,
        data_size: int,
        label_pos: int,
        label_size: int,
    ) -> None:
        self._index_writer.write(  # type: ignore
            dict(
                id=idx,
                batch=self._batch,
                data=dict(
                    file=self._DATA_FMT.format(index=fno),
                    offset=data_pos,
                    size=data_size,
                ),
                label=dict(
                    file=self._LABEL_FMT.format(index=fno),
                    offset=label_pos,
                    size=label_size,
                ),
            )
        )

    def make_swds(self) -> None:
        # TODO: add lock
        fno, wrote_size = 0, 0
        dwriter = (self.output_dir / self._DATA_FMT.format(index=fno)).open("wb")
        lwriter = (self.output_dir / self._LABEL_FMT.format(index=fno)).open("wb")

        for idx, (data, label) in enumerate(
            zip(self.iter_all_dataset_slice(), self.iter_all_label_slice())
        ):
            data_pos, data_size = self._write(dwriter, idx, data)
            label_pos, label_size = self._write(lwriter, idx, label)
            self._write_index(idx, fno, data_pos, data_size, label_pos, label_size)

            wrote_size += data_size
            if wrote_size > self.volume_bytes_size:
                wrote_size = 0
                fno += 1

                dwriter.close()
                lwriter.close()

                dwriter = (self.output_dir / self._DATA_FMT.format(index=fno)).open(
                    "wb"
                )
                lwriter = (self.output_dir / self._LABEL_FMT.format(index=fno)).open(
                    "wb"
                )

        try:
            dwriter.close()
            lwriter.close()
        except Exception as e:
            print(f"data/label write close exception: {e}")

    def _iter_files(
        self, filter: str, sort_key: t.Optional[t.Any] = None
    ) -> t.Generator[Path, None, None]:
        _key = sort_key
        if _key is not None and not callable(_key):
            raise Exception(f"data_sort_func({_key}) is not callable.")

        _files = sorted(self.data_dir.rglob(filter), key=_key)
        for p in _files:
            if not p.is_file():
                continue
            yield p

    def iter_data_files(self) -> t.Generator[Path, None, None]:
        return self._iter_files(self.data_filter, self.data_sort_func())

    def iter_label_files(self) -> t.Generator[Path, None, None]:
        return self._iter_files(self.label_filter, self.label_sort_func())

    def iter_all_dataset_slice(self) -> t.Generator[t.Any, None, None]:
        for p in self.iter_data_files():
            for d in self.iter_data_slice(str(p.absolute())):
                yield d

    def iter_all_label_slice(self) -> t.Generator[t.Any, None, None]:
        for p in self.iter_label_files():
            for d in self.iter_label_slice(str(p.absolute())):
                yield d

    @abstractmethod
    def iter_data_slice(self, path: str) -> t.Generator[t.Any, None, None]:
        raise NotImplementedError

    @abstractmethod
    def iter_label_slice(self, path: str) -> t.Generator[t.Any, None, None]:
        raise NotImplementedError

    def data_sort_func(self) -> t.Any:
        return None

    def label_sort_func(self) -> t.Any:
        return None


class MNISTBuildExecutor(BuildExecutor):
    def iter_data_slice(self, path: str):
        fpath = Path(path)

        with fpath.open("rb") as f:
            _, number, height, width = struct.unpack(">IIII", f.read(16))
            print(f">data({fpath.name}) split {math.ceil(number / self._batch)} group")

            while True:
                content = f.read(self._batch * height * width)
                if not content:
                    break
                yield content

    def iter_label_slice(self, path: str):
        fpath = Path(path)

        with fpath.open("rb") as f:
            _, number = struct.unpack(">II", f.read(8))
            print(f">label({fpath.name}) split {math.ceil(number / self._batch)} group")

            while True:
                content = f.read(self._batch)
                if not content:
                    break
                yield content


# TODO: define some open dataset class, like ImageNet, COCO



from __future__ import annotations

import json
import typing as t
import tarfile
from abc import ABCMeta
from pathlib import Path
from collections import defaultdict

import yaml
from fs import open_fs
from loguru import logger
from fs.copy import copy_fs, copy_file

from starwhale.utils import console
from starwhale.consts import (
    JSON_INDENT,
    DataLoaderKind,
    DefaultYAMLName,
    SWDSBackendType,
    SWDSSubFileType,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    VERSION_PREFIX_CNT,
    SWDS_DATA_FNAME_FMT,
    DEFAULT_COPY_WORKERS,
    LOCAL_FUSE_JSON_NAME,
    SWDS_LABEL_FNAME_FMT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import (
    move_dir,
    ensure_dir,
    ensure_file,
    blake2b_file,
    BLAKE2B_SIGNATURE_ALGO,
)
from starwhale.base.type import URIType, BundleType, InstanceType
from starwhale.base.cloud import CloudRequestMixed, CloudBundleModelMixin
from starwhale.utils.http import ignore_error
from starwhale.utils.load import import_cls
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import ExistedError, NoSupportError
from starwhale.utils.progress import run_with_progress_bar
from starwhale.base.bundle_copy import BundleCopy

from .store import DatasetStorage
from .dataset import DatasetConfig, DSProcessMode, ARCHIVE_SWDS_META


class Dataset(BaseBundle):
    __metaclass__ = ABCMeta

    def __str__(self) -> str:
        return f"Starwhale Dataset: {self.uri}"

    @classmethod
    def render_fuse_json(cls, workdir: Path, force: bool = False) -> str:
        return StandaloneDataset.render_fuse_json(workdir, force)

    @classmethod
    def get_dataset(cls, uri: URI) -> Dataset:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        bc = BundleCopy(src_uri, dest_uri, URIType.DATASET, force)
        bc.do()

    @classmethod
    def _get_cls(  # type: ignore
        cls, uri: URI
    ) -> t.Union[t.Type[StandaloneDataset], t.Type[CloudDataset]]:
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneDataset
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudDataset
        else:
            raise NoSupportError(f"dataset uri:{uri}")


class StandaloneDataset(Dataset, LocalStorageBundleMixin):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.store = DatasetStorage(uri)
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[str, t.Any] = {}  # TODO: use manifest classget_conda_env

    def add_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.add(tags, quiet)

    def remove_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.remove(tags, quiet)

    @classmethod
    def render_fuse_json(cls, workdir: Path, force: bool = False) -> str:
        _mf = workdir / DEFAULT_MANIFEST_NAME
        if not _mf.exists():
            raise Exception(f"need {DEFAULT_MANIFEST_NAME} @ {workdir}")

        _manifest = yaml.safe_load(_mf.open())
        _fuse = dict(
            backend=SWDSBackendType.FUSE,
            kind=DataLoaderKind.SWDS,
            swds=[],
        )

        ds_name = _manifest["name"]
        ds_version = _manifest["version"]
        swds_bins = [
            _k
            for _k in _manifest["signature"].keys()
            if _k.startswith("data_") and _k.endswith(SWDSSubFileType.BIN)
        ]

        bucket = workdir.parent.parent.parent
        path_prefix = f"{ds_name}/{ds_version[:VERSION_PREFIX_CNT]}/{ds_version}{BundleType.DATASET}/data"
        for idx in range(0, len(swds_bins)):
            _fuse["swds"].append(  # type: ignore
                dict(
                    bucket=str(bucket.resolve()),
                    key=dict(
                        data=f"{path_prefix}/{SWDS_DATA_FNAME_FMT.format(index=idx)}",
                        label=f"{path_prefix}/{SWDS_LABEL_FNAME_FMT.format(index=idx)}",
                        # TODO: add extra_attr ds_name, ds_versoin
                    ),
                    ext_attr=dict(
                        ds_name=ds_name,
                        ds_version=ds_version,
                    ),
                )
            )

        _f = workdir / LOCAL_FUSE_JSON_NAME
        if _f.exists() and not force:
            console.print(f":joy_cat: {LOCAL_FUSE_JSON_NAME} existed, skip render")
        else:
            ensure_file(_f, json.dumps(_fuse, indent=JSON_INDENT))
            console.print(
                f":clap: render swds {ds_name}:{ds_version} {LOCAL_FUSE_JSON_NAME}"
            )

        console.print(f":mag: {_f}")
        return str(_f.resolve())

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _r = []

        for _bf in self.store.iter_bundle_history():
            _manifest = yaml.safe_load((_bf.path / DEFAULT_MANIFEST_NAME).open())

            _r.append(
                dict(
                    name=_manifest["name"],
                    version=_bf.version,
                    size=_manifest.get("dataset_byte_size", 0),
                    created_at=_manifest["created_at"],
                    tags=_bf.tags,
                    path=_bf.path,
                )
            )

        return _r, {}

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: remove by tag
        return move_dir(self.store.snapshot_workdir, self.store.recover_loc, force)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        dest_path = (
            self.store.bundle_dir / f"{self.uri.object.version}{BundleType.DATASET}"
        )
        return move_dir(self.store.recover_loc, dest_path, force)

    def info(self) -> t.Dict[str, t.Any]:
        return self._get_bundle_info()

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        rs = defaultdict(list)

        for _bf in DatasetStorage.iter_all_bundles(
            project_uri,
            bundle_type=BundleType.DATASET,
            uri_type=URIType.DATASET,
        ):
            _manifest = yaml.safe_load((_bf.path / DEFAULT_MANIFEST_NAME).open())

            rs[_bf.name].append(
                dict(
                    name=_manifest["name"],
                    version=_bf.version,
                    size=_manifest.get("dataset_byte_size", 0),
                    created_at=_manifest["created_at"],
                    is_removed=_bf.is_removed,
                    path=_bf.path,
                    tags=_bf.tags,
                )
            )

        return rs, {}

    def build(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.DATASET, **kw: t.Any
    ) -> None:
        _dp = workdir / yaml_name
        _dataset_config = self._load_dataset_config(_dp)

        # TODO: design dataset layer mechanism
        # TODO: design append some new data into existed dataset
        # TODO: design uniq build steps for model build, swmp build

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_snapshot, 5, "prepare snapshot"),
            (
                self._copy_src,
                15,
                "copy src",
                dict(
                    workdir=workdir,
                    yaml_name=yaml_name,
                    pkg_data=_dataset_config.pkg_data,
                    exclude_pkg_data=_dataset_config.exclude_pkg_data,
                ),
            ),
            (
                self._call_make_swds,
                30,
                "make swds",
                dict(workdir=workdir, swds_config=_dataset_config),
            ),
            (self._calculate_signature, 5, "calculate signature"),
            (
                self._render_manifest,
                5,
                "render manifest",
                dict(user_raw_config=_dataset_config.as_dict()),
            ),
            (self._make_swds_meta_tar, 15, "make meta tar"),
            (self._make_latest_tag, 5, "make latest tag"),
        ]
        run_with_progress_bar("swds building...", operations)

    def _call_make_swds(self, workdir: Path, swds_config: DatasetConfig) -> None:
        from starwhale.api._impl.dataset import BuildExecutor

        logger.info("[step:swds]try to gen swds...")
        self._manifest["dataset_attr"] = swds_config.attr.as_dict()
        self._manifest["mode"] = swds_config.mode
        self._manifest["process"] = swds_config.process

        # TODO: add more import format support, current is module:class
        logger.info(f"[info:swds]try to import {swds_config.process} @ {workdir}")
        _cls = import_cls(workdir, swds_config.process, BuildExecutor)

        with _cls(
            data_dir=workdir / swds_config.data_dir,
            output_dir=self.store.data_dir,
            data_filter=swds_config.data_filter,
            label_filter=swds_config.label_filter,
            batch=swds_config.attr.batch_size,
            alignment_bytes_size=swds_config.attr.alignment_size,
            volume_bytes_size=swds_config.attr.volume_size,
        ) as _obj:
            console.print(
                f":ghost: import [red]{swds_config.process}@{workdir.resolve()}[/] to make swds..."
            )
            if swds_config.mode == DSProcessMode.GENERATE:
                logger.info("[info:swds]do make swds_bin job...")
                _obj.make_swds()
            else:
                logger.info("[info:swds]skip make swds_bin")

        console.print(f"[step:swds]finish gen swds @ {self.store.data_dir}")

    def _calculate_signature(self) -> None:
        _algo = BLAKE2B_SIGNATURE_ALGO
        _sign = dict()
        total_size = 0

        logger.info(
            f"[step:signature]try to calculate signature with {_algo} @ {self.store.data_dir}"
        )
        console.print(":robot: calculate signature...")

        # TODO: _cal(self._snapshot_workdir / ARCHIVE_SWDS_META) # add meta sign into _manifest.yaml
        for f in self.store.data_dir.iterdir():
            if not f.is_file():
                continue

            _size = f.stat().st_size
            total_size += _size
            _sign[f.name] = f"{_size}:{_algo}:{blake2b_file(f)}"

        self._manifest["dataset_byte_size"] = total_size
        self._manifest["signature"] = _sign
        logger.info(
            f"[step:signature]finish calculate signature with {_algo} for {len(_sign)} files"
        )

    def _make_swds_meta_tar(self) -> None:
        out = self.store.snapshot_workdir / ARCHIVE_SWDS_META
        logger.info(f"[step:tar]try to tar for swmp meta(NOT INCLUDE DATASET){out}")
        with tarfile.open(out, "w:") as tar:
            tar.add(str(self.store.src_dir), arcname="src")
            tar.add(str(self.store.snapshot_workdir / DEFAULT_MANIFEST_NAME))
            tar.add(str(self.store.snapshot_workdir / DefaultYAMLName.DATASET))

        console.print(
            ":hibiscus: congratulation! you can run "
            f"[red bold blink] swcli dataset info {self.name}/version/{self._version}[/]"
        )

    def _prepare_snapshot(self) -> None:
        if self.store.snapshot_workdir.exists():
            raise ExistedError(str(self.store.snapshot_workdir))

        ensure_dir(self.store.data_dir)
        ensure_dir(self.store.src_dir)

        console.print(
            f":file_folder: swds workdir: [underline]{self.store.snapshot_workdir}[/]"
        )

    def _copy_src(
        self,
        workdir: Path,
        yaml_name: str,
        pkg_data: t.List[str],
        exclude_pkg_data: t.List[str],
    ) -> None:
        logger.info(f"[step:copy]start to copy src {workdir} -> {self.store.src_dir}")
        console.print(":thumbs_up: try to copy source code files...")
        workdir_fs = open_fs(str(workdir.absolute()))
        src_fs = open_fs(str(self.store.src_dir.absolute()))
        snapshot_fs = open_fs(str(self.store.snapshot_workdir.absolute()))

        copy_file(workdir_fs, yaml_name, src_fs, DefaultYAMLName.DATASET)
        copy_file(workdir_fs, yaml_name, snapshot_fs, DefaultYAMLName.DATASET)
        # TODO: tune copy src
        copy_fs(
            workdir_fs,
            src_fs,
            walker=self._get_src_walker(workdir, pkg_data, exclude_pkg_data),
            workers=DEFAULT_COPY_WORKERS,
        )

        logger.info("[step:copy]finish copy files")

    def _load_dataset_config(self, yaml_path: Path) -> DatasetConfig:
        self._do_validate_yaml(yaml_path)
        _config = DatasetConfig.create_by_yaml(yaml_path)

        if not (yaml_path.parent / _config.data_dir).exists():
            raise FileNotFoundError(f"dataset datadir:{_config.data_dir}")

        return _config


class CloudDataset(CloudBundleModelMixin, Dataset):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(project_uri, URIType.DATASET, page, size)



import sys
import typing as t

from rich import box
from rich.tree import Tree
from rich.table import Table
from rich.pretty import Pretty

from starwhale.utils import console
from starwhale.consts import (
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, EvalTaskType, InstanceType, JobOperationType
from starwhale.base.view import BaseTermView

from .model import Job


class JobTermView(BaseTermView):
    def __init__(self, job_uri: str) -> None:
        super().__init__()
        self.raw_uri = job_uri
        self.uri = URI(job_uri, expected_type=URIType.JOB)
        self.job = Job.get_job(self.uri)
        self._action_run_map = {
            JobOperationType.CANCEL: self.job.cancel,
            JobOperationType.RESUME: self.job.resume,
            JobOperationType.REMOVE: self.job.remove,
            JobOperationType.RECOVER: self.job.recover,
            JobOperationType.PAUSE: self.job.pause,
        }

    def recover(self, force: bool = False) -> None:
        self._do_action(JobOperationType.RECOVER, force)

    def remove(self, force: bool = False) -> None:
        self._do_action(JobOperationType.REMOVE, force)

    def cancel(self, force: bool = False) -> None:
        self._do_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> None:
        self._do_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> None:
        self._do_action(JobOperationType.PAUSE, force)

    @BaseTermView._simple_action_print
    def _do_action(self, action: str, force: bool = False) -> t.Tuple[bool, str]:
        return self._action_run_map[action](force)

    @BaseTermView._header
    def compare(self, job_uris: t.List[str]) -> None:
        if self.uri.instance_type != InstanceType.STANDALONE:
            console.print(
                ":dragon_face: Today job compare only works for standalone instance"
            )
            sys.exit(1)

        jobs = []
        for _u in job_uris:
            _uri = URI(_u, expected_type=URIType.JOB)
            jobs.append(Job.get_job(_uri))

        rt = self.job.compare(jobs)
        table = Table(
            box=box.ASCII,
            title=f":leafy_green: {rt['kind'].upper()} Job Compare Summary",
            caption=f":cactus: base job:{rt['base']['uri']}",
            expand=True,
        )
        table.add_column("", justify="left", no_wrap=True, style="cyan")
        for _v in rt["versions"]:
            table.add_column(_v[:SHORT_VERSION_CNT])

        for _k, _vs in rt["summary"].items():
            _ts = [_k]

            _is_empty = all([_v["value"] is None for _v in _vs])
            if _is_empty:
                continue

            for _v in _vs:
                if _v["base"]:
                    _ts.append(f":triangular_flag: {_v['value']}")
                else:
                    if _v["delta"] is None:
                        _ts.append(_v["value"])
                    else:
                        if _v["delta"] > 0:
                            _text = (
                                f":chart_with_upwards_trend: [green]{_v['value']}[/]"
                            )
                        elif _v["delta"] < 0:
                            _text = (
                                f":chart_with_downwards_trend: [red]{_v['value']}[/]"
                            )
                        else:
                            _text = f":white_circle: {_v['value']}"
                        _ts.append(_text)

            table.add_row(*([str(_t) for _t in _ts]))

        console.print(table)

    @BaseTermView._header
    def info(self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE) -> None:
        _rt = self.job.info(page, size)
        if not _rt:
            console.print(":tea: not found info")
            return

        if _rt.get("manifest"):
            console.rule(
                f"[green bold]Inspect {DEFAULT_MANIFEST_NAME} for eval:{self.uri}"
            )
            console.print(Pretty(_rt["manifest"], expand_all=True))

        if "location" in _rt:
            console.rule("Evaluation process dirs")
            console.print(f":cactus: ppl: {_rt['location']['ppl']}")
            console.print(f":camel: cmp: {_rt['location']['cmp']}")

        if "tasks" in _rt:
            self._print_tasks(_rt["tasks"][0])

        if "report" in _rt:
            self._render_job_report(_rt["report"])

    def _print_tasks(self, tasks: t.List[t.Dict[str, t.Any]]) -> None:
        table = Table(box=box.SIMPLE)
        table.add_column("ID", justify="left", style="cyan", no_wrap=True)
        table.add_column("UUID")
        table.add_column("Status", style="magenta")
        table.add_column("Agent")
        table.add_column("Duration")
        table.add_column("Created")
        table.add_column("Finished")

        for _t in tasks:
            status, style, icon = self.pretty_status(_t["taskStatus"])
            table.add_row(
                _t["id"],
                _t["uuid"],
                f"[{style}]{icon}{status}[/]",
                _t["agent"]["ip"],
                "",
                _t["created_at"],
                "",
            )

        console.rule(
            f"[bold green]Project({self.uri.project} Job({self.job.name}) Tasks List"
        )
        console.print(table)

    def _render_job_report(self, report: t.Dict[str, t.Any]) -> None:
        if not report:
            console.print(":turtle: no report")
            return

        labels: t.Dict[str, t.Any] = report.get("labels", {})
        sort_label_names = sorted(list(labels.keys()))

        def _print_report() -> None:
            # TODO: add other kind report
            def _r(_tree: t.Any, _obj: t.Any) -> None:
                if not isinstance(_obj, dict):
                    _tree.add(str(_obj))

                for _k, _v in _obj.items():
                    if isinstance(_v, (list, tuple)):
                        _k = f"{_k}: [green]{'|'.join(_v)}"
                    elif isinstance(_v, dict):
                        _k = _k
                    else:
                        _k = f"{_k}: [green]{_v:.4f}"

                    _ntree = _tree.add(_k)
                    if isinstance(_v, dict):
                        _r(_ntree, _v)

            tree = Tree("Summary")
            _r(tree, report["summary"])
            if len(labels) == 0:
                return

            table = Table(box=box.SIMPLE)
            table.add_column("Label", style="cyan")
            keys = labels[sort_label_names[0]]
            for _k in keys:
                table.add_column(_k.capitalize())

            for _k, _v in labels.items():
                table.add_row(_k, *(f"{_v[_k2]:.4f}" for _k2 in keys))

            console.rule(f"[bold green]{report['kind'].upper()} Report")
            console.print(self.comparison(tree, table))

        def _print_confusion_matrix() -> None:
            cm = report.get("confusion_matrix", {})
            if not cm:
                return

            btable = Table(box=box.SIMPLE)
            btable.add_column("", style="cyan")
            for n in sort_label_names:
                btable.add_column(n)
            for idx, bl in enumerate(cm.get("binarylabel", [])):
                btable.add_row(sort_label_names[idx], *[f"{_:.4f}" for _ in bl])

            mtable = Table(box=box.SIMPLE)
            mtable.add_column("Label", style="cyan")
            for n in ("TP", "TN", "FP", "FN"):
                mtable.add_column(n)
            for idx, ml in enumerate(cm.get("multilabel", [])):
                mtable.add_row(sort_label_names[idx], *[str(_) for _ in ml[0] + ml[1]])

            console.rule(f"[bold green]{report['kind'].upper()} Confusion Matrix")
            console.print(self.comparison(mtable, btable))

        _print_report()
        _print_confusion_matrix()

    @classmethod
    def create(
        cls,
        project_uri: str,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        name: str = "",
        desc: str = "",
        resource: str = "",
        gencmd: bool = False,
        use_docker: bool = False,
        phase: str = EvalTaskType.ALL,
    ) -> None:
        _project_uri = URI(project_uri, expected_type=URIType.PROJECT)
        ok, reason = Job.create(
            _project_uri,
            model_uri,
            dataset_uris,
            runtime_uri,
            name=name,
            desc=desc,
            phase=phase,
            resource=resource,
            gencmd=gencmd,
            use_docker=use_docker,
        )

        # TODO: show report in standalone mode directly

        if ok:
            console.print(
                f":clap: success to create job(project id: [red]{_project_uri.full_uri}[/])"
            )
            if _project_uri.instance_type == InstanceType.CLOUD:
                _job_uri = f"{_project_uri.full_uri}/job/{reason}"
            else:
                _job_uri = f"{reason[:SHORT_VERSION_CNT]}"

            console.print(
                f":bird: run cmd to fetch job info: [bold green]swcli job info {_job_uri}[/]"
            )
        else:
            console.print(f":collision: failed to create job, notice: [red]{reason}[/]")

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        return Job.list(URI(project_uri, expected_type=URIType.PROJECT), page, size)


class JobTermViewRich(JobTermView):
    @classmethod
    @BaseTermView._pager
    @BaseTermView._header
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        _jobs, _pager = super().list(project_uri, fullname, show_removed, page, size)
        table = Table(title="Job List", box=box.SIMPLE)
        table.add_column("Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Model", no_wrap=True)
        table.add_column("Datasets")
        table.add_column("Runtime")
        table.add_column("Status", style="red")
        table.add_column("Resource", style="blue")
        table.add_column("Created At", style="magenta")
        table.add_column("Finished At", style="magenta")

        def _s(x: str) -> str:
            _end = -1 if fullname else SHORT_VERSION_CNT
            if ":" in x:
                _n, _v = x.split(":")
                return f"{_n}:{_v[:_end]}"
            else:
                return x[:_end]

        for _job in _jobs:
            if show_removed ^ _job.get("is_removed", False):
                continue

            _m = _job["manifest"]
            _status, _style, _icon = cls.pretty_status(
                _m.get("jobStatus") or _m.get("status")
            )
            _datasets = "--"
            if _m.get("datasets"):
                _datasets = "\n".join([_s(d) for d in _m["datasets"]])

            _model = "--"
            if "model" in _m:
                _model = _s(_m["model"])
            else:
                _model = f"{_m['modelName']}:{_s(_m['modelVersion'])}"

            _name = "--"
            if "id" in _m:
                _name = _m["id"]
            else:
                _name = _m["version"] if show_removed else _s(_m["version"])

            _runtime = "--"
            if "runtime" in _m:
                if isinstance(_m["runtime"], str):
                    _runtime = _m["runtime"]
                else:
                    _r = _m["runtime"]
                    _runtime = f"[{_r['version']['id']}] {_r['name']}:{_r['version']['name'][:SHORT_VERSION_CNT]}"

            table.add_row(
                _name,
                _model,
                _datasets,
                _runtime,
                f"[{_style}]{_icon}{_status}[/]",
                f"{_m['device']}:{_m['deviceAmount']}" if "device" in _m else "--",
                _m["created_at"],
                _m["finished_at"],
            )
            # TODO: add duration
        console.print(table)
        return _jobs, _pager


class JobTermViewJson(JobTermView):
    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        _data, _ = super().list(project_uri, fullname, show_removed, page, size)
        cls.pretty_json(_data)

    def info(self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE) -> None:
        _rt = self.job.info(page, size)
        if not _rt:
            console.print(":tea: not found info")
            return
        self.pretty_json(_rt)


def get_term_view(ctx_obj: t.Dict) -> t.Type[JobTermView]:
    return JobTermViewJson if ctx_obj.get("output") == "json" else JobTermViewRich



import typing as t

import click

from starwhale.base.type import EvalTaskType

from .view import JobTermView, get_term_view, DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE


@click.group("job", help="Job management, create/list/info/compare evaluation job")
@click.pass_context
def job_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@job_cmd.command("list", help="List all jobs in the current project")
@click.option("-p", "--project", default="", help="Project URI")
@click.option("--fullname", is_flag=True, help="Show fullname of swmp version")
@click.option("--show-removed", is_flag=True, help="Show removed dataset")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for job list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for job list"
)
@click.pass_obj
def _list(
    view: t.Type[JobTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
) -> None:
    view.list(
        project, fullname=fullname, show_removed=show_removed, page=page, size=size
    )


@job_cmd.command("create", help="Create job")
@click.argument("project", default="")
@click.option("--model", required=True, help="model uri or model.yaml dir path")
@click.option(
    "--dataset",
    required=True,
    multiple=True,
    help="dataset uri, one or more",
)
@click.option("--runtime", default="", help="runtime uri")
@click.option("--name", help="job name")
@click.option("--desc", help="job description")
@click.option(
    "--resource",
    default="cpu:1",
    type=str,
    help="[ONLY Cloud]resource, fmt is resource [name]:[cnt], such as cpu:1, gpu:2",
)
@click.option(
    "--use-docker",
    is_flag=True,
    help="[ONLY Standalone]use docker to run evaluation job",
)
@click.option("--gencmd", is_flag=True, help="[ONLY Standalone]gen docker run command")
@click.option(
    "--phase",
    type=click.Choice([EvalTaskType.ALL, EvalTaskType.PPL]),
    default=EvalTaskType.ALL,
    help="[ONLY Standalone]evaluation run phase",
)
def _create(
    project: str,
    model: str,
    dataset: t.List[str],
    runtime: str,
    name: str,
    desc: str,
    resource: str,
    use_docker: bool,
    gencmd: bool,
    phase: str,
) -> None:
    JobTermView.create(
        project_uri=project,
        model_uri=model,
        dataset_uris=dataset,
        runtime_uri=runtime,
        name=name,
        desc=desc,
        resource=resource,
        gencmd=gencmd,
        phase=phase,
        use_docker=use_docker,
    )


@job_cmd.command("remove", help="Remove job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to remove")
def _remove(job: str, force: bool) -> None:
    click.confirm("continue to remove?", abort=True)
    JobTermView(job).remove(force)


@job_cmd.command("recover", help="Recover removed job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to recover")
def _recover(job: str, force: bool) -> None:
    JobTermView(job).recover(force)


@job_cmd.command("pause", help="Pause job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to pause")
def _pause(job: str, force: bool) -> None:
    click.confirm("continue to pause?", abort=True)
    JobTermView(job).pause(force)


@job_cmd.command("resume", help="Resume job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to resume")
def _resume(job: str, force: bool) -> None:
    JobTermView(job).resume(force)


@job_cmd.command("cancel", help="Cancel job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to cancel")
def _cancel(job: str, force: bool) -> None:
    click.confirm("continue to cancel?", abort=True)
    JobTermView(job).cancel(force)


@job_cmd.command("info", help="Inspect job details")
@click.argument("job")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for tasks list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for tasks list"
)
@click.pass_obj
def _info(view: t.Type[JobTermView], job: str, page: int, size: int) -> None:
    view(job).info(page, size)


@job_cmd.command("compare")
@click.argument("base_job", nargs=1)
@click.argument("job", nargs=-1)
def _compare(base_job: str, job: t.List[str]) -> None:
    """
    [ONLY Standalone]Compare the result of evaluation job

    BASE_JOB: job uri

    JOB: job uri
    """
    JobTermView(base_job).compare(job)



class SWEnv:
    task_id = "SW_TASK_ID"
    job_id = "SW_JOB_ID"
    status_dir = "SW_TASK_STATUS_DIR"
    log_dir = "SW_TASK_LOG_DIR"
    result_dir = "SW_TASK_RESULT_DIR"
    input_config = "SW_TASK_INPUT_CONFIG"



from __future__ import annotations

import typing as t
from copy import deepcopy
from pathlib import Path

import yaml

from starwhale.utils import convert_to_bytes
from starwhale.consts import SWDSSubFileType, DEFAULT_STARWHALE_API_VERSION
from starwhale.utils.error import NoSupportError


class DSProcessMode:
    DEFINE = "define"
    GENERATE = "generate"


D_FILE_VOLUME_SIZE = 64 * 1024 * 1024  # 64MB
D_ALIGNMENT_SIZE = 4 * 1024  # 4k for page cache
D_USER_BATCH_SIZE = 1
ARCHIVE_SWDS_META = "archive.%s" % SWDSSubFileType.META


# TODO: use attr to tune code
class DatasetAttr(object):
    def __init__(
        self,
        volume_size: t.Union[int, str] = D_FILE_VOLUME_SIZE,
        alignment_size: t.Union[int, str] = D_ALIGNMENT_SIZE,
        batch_size: int = D_USER_BATCH_SIZE,
        **kw: t.Any,
    ) -> None:
        self.batch_size = batch_size
        self.volume_size = convert_to_bytes(volume_size)
        self.alignment_size = convert_to_bytes(alignment_size)
        self.kw = kw

    def as_dict(self) -> t.Dict[str, t.Any]:
        _rd = deepcopy(self.__dict__)
        _rd.pop("kw")
        return _rd


# TODO: abstract base class from DataSetConfig and ModelConfig
# TODO: use attr to tune code
class DatasetConfig(object):
    def __init__(
        self,
        name: str,
        data_dir: str,
        process: str,
        mode: str = DSProcessMode.GENERATE,
        data_filter: str = "",
        label_filter: str = "",
        runtime: str = "",
        pkg_data: t.List[str] = [],
        exclude_pkg_data: t.List[str] = [],
        tag: t.List[str] = [],
        desc: str = "",
        version: str = DEFAULT_STARWHALE_API_VERSION,
        attr: t.Dict[str, t.Any] = {},
        **kw: t.Any,
    ) -> None:
        self.name = name
        self.mode = mode
        self.data_dir = str(data_dir)
        self.data_filter = data_filter
        self.label_filter = label_filter
        self.process = process
        self.tag = tag
        self.desc = desc
        self.version = version
        self.runtime = runtime.strip()
        self.attr = DatasetAttr(**attr)
        self.pkg_data = pkg_data
        self.exclude_pkg_data = exclude_pkg_data
        self.kw = kw

        self._validator()

    def _validator(self) -> None:
        if self.mode not in (DSProcessMode.DEFINE, DSProcessMode.GENERATE):
            raise NoSupportError(f"{self.mode} mode no support")

        if ":" not in self.process:
            raise Exception(
                f"please use module:class format, current is: {self.process}"
            )

        # TODO: add more validator

    def __str__(self) -> str:
        return f"DataSet Config {self.name}"

    def __repr__(self) -> str:
        return f"DataSet Config {self.name}, mode:{self.mode}, data:{self.data_dir}"

    def as_dict(self) -> t.Dict[str, t.Any]:
        _r = deepcopy(self.__dict__)
        _r["attr"] = self.attr.as_dict()
        return _r

    @classmethod
    def create_by_yaml(cls, fpath: t.Union[str, Path]) -> DatasetConfig:
        fpath = Path(fpath)

        with fpath.open("r") as f:
            c = yaml.safe_load(f)

        return cls(**c)



from __future__ import annotations

import typing as t
from functools import wraps

from sklearn.metrics import (  # type: ignore
    auc,
    roc_curve,
    hamming_loss,
    confusion_matrix,
    cohen_kappa_score,
    classification_report,
    multilabel_confusion_matrix,
)


class MetricKind:
    MultiClassification = "multi_classification"


def multi_classification(
    confusion_matrix_normalize: str = "all",
    show_hamming_loss: bool = True,
    show_cohen_kappa_score: bool = True,
    show_roc_auc: bool = True,
    all_labels: t.Optional[t.List[t.Any]] = None,
) -> t.Any:
    def _decorator(func: t.Any) -> t.Any:
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> t.Dict[str, t.Any]:
            y_pr: t.Any = None

            _rt = func(*args, **kwargs)
            if show_roc_auc:
                y_true, y_pred, y_pr = _rt
            else:
                y_true, y_pred = _rt

            _r: t.Dict[str, t.Any] = {"kind": MetricKind.MultiClassification}
            cr = classification_report(
                y_true, y_pred, output_dict=True, labels=all_labels
            )
            _summary_m = ["accuracy", "macro avg", "weighted avg"]
            _r["summary"] = {k: cr.get(k) for k in _summary_m}
            _r["labels"] = {k: v for k, v in cr.items() if k not in _summary_m}

            # TODO: tune performace, use intermediated result
            cm = confusion_matrix(
                y_true, y_pred, labels=all_labels, normalize=confusion_matrix_normalize
            )
            mcm = multilabel_confusion_matrix(y_true, y_pred, labels=all_labels)
            _r["confusion_matrix"] = {
                "binarylabel": cm.tolist(),
                "multilabel": mcm.tolist(),
            }
            if show_hamming_loss:
                _r["summary"]["hamming_loss"] = hamming_loss(y_true, y_pred)
            if show_cohen_kappa_score:
                _r["summary"]["cohen_kappa_score"] = cohen_kappa_score(y_true, y_pred)

            if show_roc_auc and all_labels is not None and y_true and y_pr:
                _r["roc_auc"] = {}
                for _idx, _label in enumerate(all_labels):
                    _r["roc_auc"][_label] = _calculate_roc_auc(
                        y_true, y_pr, _label, _idx
                    )
            return _r

        return _wrapper

    return _decorator


def _calculate_roc_auc(
    y_pred: t.List[t.Any],
    y_pr: t.List[t.Any],
    label: t.Any,
    idx: int,
) -> t.Dict[str, t.Any]:
    # TODO: add sample rate
    y_bin_true = [int(i == label) for i in y_pred]
    y_label_pr = [m[idx] for m in y_pr]
    fpr, tpr, thresholds = roc_curve(y_bin_true, y_label_pr)
    return dict(
        fpr=fpr.tolist(),
        tpr=tpr.tolist(),
        auc=auc(fpr, tpr),
        thresholds=thresholds.tolist(),
    )



import os
import typing as t
from select import select
from subprocess import PIPE, Popen, STDOUT, CalledProcessError

from loguru import logger


def log_check_call(*args: t.Any, **kwargs: t.Any) -> int:
    log = kwargs.pop("log", logger.debug)
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = STDOUT
    env = kwargs.get("env", os.environ.copy())
    kwargs["env"] = env
    kwargs["universal_newlines"] = True
    env["PYTHONUNBUFFERED"] = "1"

    output = []
    p = Popen(*args, **kwargs)
    log(f"cmd: {p.args!r}")
    while True:
        fds, _, _ = select([p.stdout], [], [], 30)  # timeout 30s
        for fd in fds:
            line = fd.readline()
            log(line.rstrip())
            output.append(line)
        else:
            if p.poll() is not None:
                break

    p.wait()

    if p.returncode != 0:
        cmd = kwargs.get("args") or args[0]
        e = CalledProcessError(p.returncode, cmd)
        e.output = "".join(output)
        raise e

    return 0


check_call = log_check_call



from http import HTTPStatus
from pathlib import Path

import yaml
from rich.progress import (
    TaskID,
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    TotalFileSizeColumn,
    TransferSpeedColumn,
)

from starwhale.utils import console
from starwhale.consts import HTTPMethod, VERSION_PREFIX_CNT, DEFAULT_MANIFEST_NAME
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir
from starwhale.base.type import URIType, InstanceType, get_bundle_type_by_uri
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.error import NotFoundError, NoSupportError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.core.model.store import ModelStorage
from starwhale.core.dataset.store import DatasetStorage
from starwhale.core.runtime.store import RuntimeStorage
from starwhale.core.dataset.dataset import ARCHIVE_SWDS_META

TMP_FILE_BUFSIZE = 8192

_query_param_map = {
    URIType.DATASET: "swds",
    URIType.MODEL: "swmp",
    URIType.RUNTIME: "runtime",
}


class _UploadPhase:
    MANIFEST = "MANIFEST"
    BLOB = "BLOB"
    END = "END"
    CANCEL = "CANCEL"


class BundleCopy(CloudRequestMixed):
    def __init__(
        self, src_uri: str, dest_uri: str, typ: str, force: bool = False
    ) -> None:
        self.src_uri = URI(src_uri, expected_type=typ)
        self.dest_uri = URI(dest_uri, expected_type=URIType.PROJECT)
        self.typ = typ
        self.force = force

        self.bundle_name = self.src_uri.object.name
        self.bundle_version = self._guess_bundle_version()

        self._sw_config = SWCliConfigMixed()
        self._do_validate()

    def _guess_bundle_version(self) -> str:
        if self.src_uri.instance_type == InstanceType.CLOUD:
            return self.src_uri.object.version
        else:
            if self.typ == URIType.DATASET:
                _v = DatasetStorage(self.src_uri).id
            elif self.typ == URIType.MODEL:
                _v = ModelStorage(self.src_uri).id
            elif self.typ == URIType.RUNTIME:
                _v = RuntimeStorage(self.src_uri).id
            else:
                raise NoSupportError(self.typ)

            return _v or self.src_uri.object.version

    def _do_validate(self) -> None:
        if self.typ not in (URIType.DATASET, URIType.MODEL, URIType.RUNTIME):
            raise NoSupportError(f"{self.typ} copy does not work")

        if self.bundle_version == "":
            raise Exception(f"must specify version src:({self.bundle_version})")

    def _check_cloud_obj_existed(self, uri: URI) -> bool:
        # TODO: add more params for project
        # TODO: tune controller api, use unified params name
        ok, _ = self.do_http_request_simple_ret(
            path=f"/project/{self.typ}",
            method=HTTPMethod.HEAD,
            instance_uri=uri,
            params={
                _query_param_map[self.typ]: f"{self.bundle_name}:{self.bundle_version}",
                "project": uri.project,
            },
            ignore_status_codes=[HTTPStatus.NOT_FOUND],
        )
        return ok

    def _get_target_path(self, uri: URI) -> Path:
        if uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError(f"{uri} to get target dir path")

        return (
            self._sw_config.rootdir
            / uri.project
            / self.typ
            / self.bundle_name
            / self.bundle_version[:VERSION_PREFIX_CNT]
            / f"{self.bundle_version}{get_bundle_type_by_uri(self.typ)}"
        )

    def _is_existed(self, uri: URI) -> bool:
        if uri.instance_type == InstanceType.CLOUD:
            return self._check_cloud_obj_existed(uri)
        else:
            return self._get_target_path(uri).exists()

    def _do_upload_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_target_path(self.src_uri)
        task_id = progress.add_task(
            f":bowling: upload {file_path.name}",
            total=file_path.stat().st_size,
        )

        self.do_multipart_upload_file(
            url_path=f"/project/{self.typ}/push",
            file_path=file_path,
            instance_uri=self.dest_uri,
            fields={
                _query_param_map[self.typ]: f"{self.bundle_name}:{self.bundle_version}",
                "project": self.dest_uri.project,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )

    def _do_download_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_target_path(self.dest_uri)
        task_id = progress.add_task(f":bowling: download to {file_path}...")

        self.do_download_file(
            url_path=f"/project/{self.typ}/pull",
            dest_path=self._get_target_path(self.dest_uri),
            instance_uri=self.src_uri,
            params={
                _query_param_map[self.typ]: f"{self.bundle_name}:{self.bundle_version}",
                "project": self.src_uri.project,
            },
            progress=progress,
            task_id=task_id,
        )

    def do(self) -> None:
        if self._is_existed(self.dest_uri) and not self.force:
            console.print(
                f":tea: {self.dest_uri}-{self.bundle_name}-{self.bundle_version} was already existed, skip copy"
            )
            return

        # TODO: when controller api support dataset head, remove dataset type check
        if self.typ != URIType.DATASET and not self._is_existed(self.src_uri):
            raise NotFoundError(str(self.src_uri))

        console.print(
            f":construction: start to copy {self.src_uri} -> {self.dest_uri}..."
        )

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeElapsedColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            console=console,
            refresh_per_second=0.2,
        ) as progress:
            if self.src_uri.instance_type == InstanceType.STANDALONE:
                if self.typ == URIType.DATASET:
                    self._do_upload_bundle_dir(progress)
                else:
                    self._do_upload_bundle_tar(progress)
            else:
                if self.typ == URIType.DATASET:
                    self._do_download_bundle_dir(progress)
                else:
                    self._do_download_bundle_tar(progress)

    def _do_upload_bundle_dir(self, progress: Progress) -> None:
        _workdir: Path = self._get_target_path(self.src_uri)
        _manifest_path = _workdir / DEFAULT_MANIFEST_NAME
        _key = _query_param_map[self.typ]
        _name = f"{self.bundle_name}:{self.bundle_version}"
        _url_path = f"/project/{self.typ}/push"

        task_id = progress.add_task(
            f":arrow_up: {_manifest_path.name}",
            total=_manifest_path.stat().st_size,
        )

        # TODO: use rich progress
        r = self.do_multipart_upload_file(
            url_path=_url_path,
            file_path=_manifest_path,
            instance_uri=self.dest_uri,
            fields={
                _key: _name,
                "phase": _UploadPhase.MANIFEST,
                "project": self.dest_uri.project,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )
        upload_id = r.json().get("data", {}).get("upload_id")
        if not upload_id:
            raise Exception("get invalid upload_id")
        _headers = {"X-SW-UPLOAD-ID": str(upload_id)}
        _manifest = yaml.safe_load(_manifest_path.open())

        # TODO: add retry deco
        def _upload_blob(_fp: Path, _tid: TaskID) -> None:
            if not _fp.exists():
                raise NotFoundError(f"{_fp} not found")

            self.do_multipart_upload_file(
                url_path=_url_path,
                file_path=_fp,
                instance_uri=self.dest_uri,
                fields={
                    _key: _name,
                    "phase": _UploadPhase.BLOB,
                },
                headers=_headers,
                use_raise=True,
                progress=progress,
                task_id=_tid,
            )

        try:
            from starwhale.core.dataset.dataset import ARCHIVE_SWDS_META

            _p_map = {}
            for _p in [_workdir / "data" / n for n in _manifest["signature"]] + [
                _workdir / ARCHIVE_SWDS_META
            ]:
                _tid = progress.add_task(
                    f":arrow_up: {_p.name}",
                    total=_p.stat().st_size,
                )
                _p_map[_tid] = _p

            for _tid, _p in _p_map.items():
                _upload_blob(_p, _tid)

        except Exception as e:
            console.print(
                f":confused_face: when upload blobs, we meet Exception{e}, will cancel upload"
            )
            self.do_http_request(
                path=_url_path,
                method=HTTPMethod.POST,
                instance_uri=self.dest_uri,
                headers=_headers,
                data={
                    _key: _name,
                    "project": self.dest_uri.project,
                    "phase": _UploadPhase.CANCEL,
                },
                use_raise=True,
                disable_default_content_type=True,
            )
        else:
            self.do_http_request(
                path=_url_path,
                method=HTTPMethod.POST,
                instance_uri=self.dest_uri,
                headers=_headers,
                data={
                    _key: _name,
                    "project": self.dest_uri.project,
                    "phase": _UploadPhase.END,
                },
                use_raise=True,
                disable_default_content_type=True,
            )

    def _do_download_bundle_dir(self, progress: Progress) -> None:
        _workdir = self._get_target_path(self.dest_uri)
        ensure_dir(_workdir)
        ensure_dir(_workdir / "data")

        def _download(_target: Path, _part: str, _tid: TaskID) -> None:
            self.do_download_file(
                # TODO: use /project/{self.typ}/pull api
                url_path=f"/project/{self.typ}/pull",
                dest_path=_target,
                instance_uri=self.src_uri,
                params={
                    "name": self.bundle_name,
                    "version": self.bundle_version,
                    "part_name": _part,
                },
                progress=progress,
                task_id=_tid,
            )

        _manifest_path = _workdir / DEFAULT_MANIFEST_NAME
        _tid = progress.add_task(f":arrow_down: {DEFAULT_MANIFEST_NAME}")
        _download(_manifest_path, DEFAULT_MANIFEST_NAME, _tid)
        _manifest = yaml.safe_load(_manifest_path.open())

        _p_map = {}
        for _k in _manifest.get("signature", {}):
            # TODO: parallel download
            _tid = progress.add_task(f":arrow_down: {_k}")
            _p_map[_tid] = {"path": _workdir / "data" / _k, "part": _k}

        _tid = progress.add_task(f":arrow_down: {ARCHIVE_SWDS_META}")
        _p_map[_tid] = {"path": _workdir / ARCHIVE_SWDS_META, "part": ARCHIVE_SWDS_META}

        for _tid, _info in _p_map.items():
            _download(_info["path"], _info["part"], _tid)



from __future__ import annotations

import io
import os
import sys
import json
import base64
import typing as t
import logging
from abc import ABCMeta, abstractmethod
from types import TracebackType
from pathlib import Path
from functools import wraps

import dill
import loguru
import jsonlines

from starwhale.utils import now_str, pretty_bytes, in_production
from starwhale.consts import CURRENT_FNAME, DEFAULT_INPUT_JSON_FNAME
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.utils.log import StreamWrapper
from starwhale.consts.env import SWEnv
from starwhale.utils.error import NotFoundError

from .loader import DataField, DataLoader, get_data_loader

_TASK_ROOT_DIR = "/var/starwhale" if in_production() else "/tmp/starwhale"

_p = lambda p, sub: Path(p) if p else Path(_TASK_ROOT_DIR) / sub
_ptype = t.Union[str, None, Path]


class _LogType:
    SW = "starwhale"
    USER = "user"


_jl_writer = lambda p: jsonlines.open(str((p).resolve()), mode="w")


class _RunConfig(object):
    def __init__(
        self,
        swds_config_path: _ptype = "",
        status_dir: _ptype = "",
        log_dir: _ptype = "",
        result_dir: _ptype = "",
    ) -> None:
        self.status_dir = _p(status_dir, "status")  # type: ignore
        self.log_dir = _p(log_dir, "log")  # type: ignore
        self.result_dir = _p(result_dir, "result")  # type: ignore
        self.swds_config = self.load_swds_config(swds_config_path)

        # TODO: graceful method
        self._prepare()

    def load_swds_config(self, path: _ptype) -> t.Any:
        if not path:
            path = Path(_TASK_ROOT_DIR) / "config" / DEFAULT_INPUT_JSON_FNAME

        path = Path(path) if isinstance(path, str) else path
        if path.exists():
            # TODO: validate swds config
            return json.load(path.open("r"))
        else:
            raise NotFoundError(f"{path} not found")

    def _prepare(self) -> None:
        ensure_dir(self.log_dir)
        ensure_dir(self.result_dir)
        ensure_dir(self.status_dir)

    @classmethod
    def create_by_env(cls) -> "_RunConfig":
        _env = os.environ.get
        return _RunConfig(
            swds_config_path=_env(SWEnv.input_config),
            status_dir=_env(SWEnv.status_dir),
            log_dir=_env(SWEnv.log_dir),
            result_dir=_env(SWEnv.result_dir),
        )

    @classmethod
    def set_env(cls, _config: t.Dict[str, t.Any] = {}) -> None:
        def _set(_k: str, _e: str) -> None:
            _v = _config.get(_k)
            if _v:
                os.environ[_e] = str(_v)

        _set("status_dir", SWEnv.status_dir)
        _set("log_dir", SWEnv.log_dir)
        _set("result_dir", SWEnv.result_dir)
        _set("input_config", SWEnv.input_config)


class PipelineHandler(object):
    class ResultOutputType:
        JSONL = "jsonline"
        PLAIN = "plain"

    class STATUS:
        START = "start"
        RUNNING = "running"
        SUCCESS = "success"
        FAILED = "failed"

    __metaclass__ = ABCMeta

    def __init__(
        self,
        merge_label: bool = True,
        output_type: str = ResultOutputType.JSONL,
        ignore_error: bool = False,
    ) -> None:
        # TODO: add args for compare result and label directly
        self.merge_label = merge_label
        self.output_type = output_type
        self.ignore_error = ignore_error
        self.config = _RunConfig.create_by_env()

        self.logger, self._sw_logger = self._init_logger()
        self._stdout_changed = False
        self._stderr_changed = False
        self._orig_stdout = sys.stdout
        self._orig_stderr = sys.stderr

        self._data_loader = get_data_loader(
            self.config.swds_config, self._sw_logger, deserializer=self.deserialize
        )
        # TODO: split status/result files
        self._result_writer = _jl_writer(self.config.result_dir / CURRENT_FNAME)  # type: ignore
        self._status_writer = _jl_writer(self.config.status_dir / "timeline")  # type: ignore

        self._ppl_data_field = "ppl"
        self._label_field = "label"

        self._monkey_patch()
        self._update_status(self.STATUS.START)

    def _init_logger(self) -> t.Tuple[loguru.Logger, loguru.Logger]:
        # TODO: remove logger first?
        # TODO: add custom log format, include daemonset pod name
        from loguru import logger as _logger

        # TODO: configure log rotation size
        _logger.add(
            self.config.log_dir / "{time}.log",
            rotation="500MB",
            backtrace=True,
            diagnose=True,
            serialize=True,
        )
        _logger.bind(
            type=_LogType.USER,
            task_id=os.environ.get("SW_TASK_ID", ""),
            job_id=os.environ.get("SW_JOB_ID", ""),
        )
        _sw_logger = _logger.bind(type=_LogType.SW)
        return _logger, _sw_logger

    def _monkey_patch(self) -> None:
        if not isinstance(sys.stdout, StreamWrapper) and isinstance(
            sys.stdout, io.TextIOWrapper
        ):
            sys.stdout = StreamWrapper(sys.stdout, self.logger, logging.INFO)  # type: ignore
            self._stdout_changed = True

        if not isinstance(sys.stderr, StreamWrapper) and isinstance(
            sys.stderr, io.TextIOWrapper
        ):
            sys.stderr = StreamWrapper(sys.stderr, self.logger, logging.WARN)  # type: ignore
            self._stderr_changed = True

    def __str__(self) -> str:
        return (
            f"PipelineHandler status@{self.config.status_dir}, "
            f"log@{self.config.log_dir}, result@{self.config.result_dir}"
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:
            print(f"type:{type}, exception:{value}, traceback:{trace}")

        if self._stdout_changed:
            sys.stdout = self._orig_stdout
        if self._stderr_changed:
            sys.stderr = self._orig_stderr

        try:
            self._result_writer.close()
        except Exception as e:
            self._sw_logger.exception(f"result writer close exception: {e}")

        try:
            self._status_writer.close()
        except Exception as e:
            self._sw_logger.exception(f"status writer close exception: {e}")

        self.logger.remove()
        self._sw_logger.remove()

    @abstractmethod
    def ppl(self, data: bytes, batch_size: int, **kw: t.Any) -> t.Any:
        # TODO: how to handle each batch element is not equal.
        raise NotImplementedError

    @abstractmethod
    def cmp(self, _data_loader: DataLoader) -> t.Any:
        raise NotImplementedError

    def _builtin_serialize(self, *data: t.Any) -> bytes:
        return dill.dumps(data)

    def ppl_data_serialize(self, *data: t.Any) -> bytes:
        return self._builtin_serialize(*data)

    def ppl_data_deserialize(self, data: bytes) -> t.Any:
        return dill.loads(base64.b64decode(data))

    def label_data_serialize(self, data: t.Any) -> bytes:
        return self._builtin_serialize(data)

    def label_data_deserialize(self, data: bytes) -> bytes:
        return dill.loads(base64.b64decode(data))[0]

    def deserialize(self, data: t.Union[str, bytes]) -> t.Any:
        ret = json.loads(data)
        ret[self._ppl_data_field] = self.ppl_data_deserialize(ret[self._ppl_data_field])
        ret[self._label_field] = self.label_data_deserialize(ret[self._label_field])
        return ret

    def handle_label(self, label: bytes, batch_size: int, **kw: t.Any) -> t.Any:
        return label.decode()

    def _record_status(func):  # type: ignore
        @wraps(func)  # type: ignore
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            self: PipelineHandler = args[0]
            self._sw_logger.info(f"start to run {func}...")
            self._update_status(self.STATUS.RUNNING)
            try:
                func(*args, **kwargs)  # type: ignore
            except Exception as e:
                self._update_status(self.STATUS.FAILED)
                self._sw_logger.exception(f"{func} abort, exception: {e}")
                raise
            else:
                self._update_status(self.STATUS.SUCCESS)
                self._sw_logger.info("finish.")

        return _wrapper

    @_record_status  # type: ignore
    def _starwhale_internal_run_cmp(self) -> None:
        now = now_str()  # type: ignore
        try:
            output = self.cmp(self._data_loader)
        except Exception as e:
            self._sw_logger.exception(f"cmp exception: {e}")
            self._status_writer.write({"time": now, "status": False, "exception": e})
            raise
        else:
            self._status_writer.write({"time": now, "status": True, "exception": ""})
            self._result_writer.write(output)

    @_record_status  # type: ignore
    def _starwhale_internal_run_ppl(self) -> None:
        for data, label in self._data_loader:
            self._sw_logger.info(
                f"[{data.idx}]data-label loaded, data size:{pretty_bytes(data.data_size)}, "
                f"label size:{pretty_bytes(label.data_size)} ,batch:{data.batch_size}"
            )

            if data.idx != label.idx:
                msg = (
                    f"data index[{data.idx}] is not equal label index [{label.idx}], "
                    f"{'ignore error' if self.ignore_error else ''}"
                )
                self._sw_logger.error(msg)
                if not self.ignore_error:
                    raise Exception(msg)

            pred: t.Any = b""
            exception = None
            try:
                # TODO: inspect profiling
                pred = self.ppl(
                    data.data,
                    data.batch_size,
                    data_index=data.idx,
                    data_size=data.data_size,
                    label_content=label.data,
                    label_size=label.data_size,
                    label_batch=label.batch_size,
                    label_index=label.idx,
                    ds_name=data.ext_attr.get("ds_name", ""),
                    ds_version=data.ext_attr.get("ds_version", ""),
                )
            except Exception as e:
                exception = e
                self._sw_logger.exception(f"[{data.idx}] data handle -> failed")
                if not self.ignore_error:
                    self._update_status(self.STATUS.FAILED)
                    raise
            else:
                exception = None
                self._sw_logger.info(f"[{data.idx}] data handle -> success")

            self._do_record(data, label, exception, *pred)

    def _do_record(
        self,
        data: DataField,
        label: DataField,
        exception: t.Union[None, Exception],
        *args,
    ) -> None:
        self._status_writer.write(
            {
                "time": now_str(),  # type: ignore
                "status": exception is None,
                "exception": str(exception),
                "index": data.idx,
                "output_tuple_len": len(args),
            }
        )

        # TODO: output maybe cannot be jsonized
        result = {
            "index": data.idx,
            self._ppl_data_field: base64.b64encode(
                self.ppl_data_serialize(*args)
            ).decode("ascii"),
            "batch": data.batch_size,
        }
        if self.merge_label:
            try:
                label = self.handle_label(
                    label.data,
                    label.batch_size,
                    index=label.idx,
                    size=label.data_size,
                )
                result[self._label_field] = base64.b64encode(
                    self.label_data_serialize(label)
                ).decode("ascii")
            except Exception as e:
                self._sw_logger.exception(f"{label.data!r} label handle exception:{e}")
                if not self.ignore_error:
                    self._update_status(self.STATUS.FAILED)
                    raise
                else:
                    result["label"] = ""

        self._result_writer.write(result)
        self._update_status(self.STATUS.RUNNING)

    def _update_status(self, status: str) -> None:
        fpath = self.config.status_dir / CURRENT_FNAME
        ensure_file(fpath, status)



from starwhale.utils.error import NoSupportError


class InstanceType:
    STANDALONE = "standalone"
    CLOUD = "cloud"


class URIType:
    INSTANCE = "instance"
    PROJECT = "project"
    MODEL = "model"
    DATASET = "dataset"
    RUNTIME = "runtime"
    JOB = "job"
    UNKNOWN = "unknown"


class JobType:
    EVAL = "evaluation"


class EvalTaskType:
    ALL = "all"
    PPL = "ppl"
    CMP = "cmp"


class RunSubDirType:
    RESULT = "result"
    DATASET = "dataset"
    PPL_RESULT = "ppl_result"
    STATUS = "status"
    LOG = "log"
    SWMP = "swmp"
    CONFIG = "config"


class JobOperationType:
    CANCEL = "cancel"
    PAUSE = "pause"
    RESUME = "resume"
    CREATE = "create"
    REMOVE = "remove"
    RECOVER = "recover"
    INFO = "info"


class BundleType:
    MODEL = ".swmp"
    DATASET = ".swds"
    RUNTIME = ".swrt"


def get_bundle_type_by_uri(uri_type: str) -> str:
    if uri_type == URIType.DATASET:
        return BundleType.DATASET
    elif uri_type == URIType.MODEL:
        return BundleType.MODEL
    elif uri_type == URIType.RUNTIME:
        return BundleType.RUNTIME
    else:
        raise NoSupportError(uri_type)



import os
import typing as t
from pathlib import Path

from starwhale.utils import console, pretty_bytes, in_production
from starwhale.consts import (
    PythonRunEnv,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    DEFAULT_PYTHON_VERSION,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.base.view import BaseTermView
from starwhale.core.runtime.store import RuntimeStorage

from .model import Runtime


class RuntimeTermView(BaseTermView):
    def __init__(self, runtime_uri: str) -> None:
        super().__init__()

        self.raw_uri = runtime_uri
        self.uri = URI(runtime_uri, expected_type=URIType.RUNTIME)
        self.runtime = Runtime.get_runtime(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.runtime.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.runtime.recover(force)

    @BaseTermView._pager
    @BaseTermView._header
    def history(
        self, fullname: bool = False
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        return self._print_history(
            title="Runtime History", history=self.runtime.history(), fullname=fullname
        )

    @BaseTermView._header
    def info(self, fullname: bool = False) -> None:
        self._print_info(self.runtime.info(), fullname=fullname)

    @classmethod
    def activate(cls, workdir: str, yaml_name: str = DefaultYAMLName.RUNTIME) -> None:
        Runtime.activate(workdir, yaml_name)

    @classmethod
    def build(
        cls,
        workdir: str,
        project: str = "",
        yaml_name: str = DefaultYAMLName.RUNTIME,
        gen_all_bundles: bool = False,
        include_editable: bool = False,
    ) -> None:
        _runtime_uri = cls.prepare_build_bundle(
            workdir, project, yaml_name, URIType.RUNTIME
        )
        if include_editable:
            console.print(
                ":bell: [red bold]runtime will include pypi editable package[/] :bell:"
            )
        else:
            console.print(
                ":bird: [red bold]runtime will ignore pypi editable package[/]"
            )

        _rt = Runtime.get_runtime(_runtime_uri)
        _rt.build(
            Path(workdir),
            yaml_name,
            gen_all_bundles=gen_all_bundles,
            include_editable=include_editable,
        )

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> None:
        console.print(":oncoming_police_car: try to extract ...")
        path = self.runtime.extract(force, target)
        console.print(f":clap: extracted @ {path.resolve()} :tada:")

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        fullname = fullname or (_uri.instance_type == InstanceType.CLOUD)
        _runtimes, _pager = Runtime.list(_uri, page, size)
        _data = BaseTermView.list_data(_runtimes, show_removed, fullname)
        return _data, _pager

    @classmethod
    def create(
        cls,
        workdir: str,
        name: str,
        python_version: str = DEFAULT_PYTHON_VERSION,
        mode: str = PythonRunEnv.VENV,
        force: bool = False,
    ) -> None:
        console.print(f":construction: start to create runtime[{name}] environment...")
        Runtime.create(
            workdir,
            name,
            python_version=python_version,
            mode=mode,
            force=force,
        )
        console.print(":clap: python runtime environment is ready to use :tada:")

    @classmethod
    def restore(cls, target: str) -> None:
        if in_production() or (os.path.exists(target) and os.path.isdir(target)):
            workdir = Path(target)
        else:
            uri = URI(target, URIType.RUNTIME)
            store = RuntimeStorage(uri)
            workdir = store.snapshot_workdir

        console.print(f":golfer: try to restore python runtime environment{workdir}...")
        Runtime.restore(workdir)

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        Runtime.copy(src_uri, dest_uri, force)
        console.print(":clap: copy done.")

    @BaseTermView._header
    def tag(self, tags: t.List[str], remove: bool = False, quiet: bool = False) -> None:
        # TODO: refactor model/runtime/dataset tag view-model
        if remove:
            console.print(f":golfer: remove tags [red]{tags}[/] @ {self.uri}...")
            self.runtime.remove_tags(tags, quiet)
        else:
            console.print(f":surfer: add tags [red]{tags}[/] @ {self.uri}...")
            self.runtime.add_tags(tags, quiet)


class RuntimeTermViewRich(RuntimeTermView):
    @classmethod
    @BaseTermView._pager
    @BaseTermView._header
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _data, _pager = super().list(project_uri, fullname, show_removed, page, size)

        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(),
        }

        cls.print_table("Runtime List", _data, custom_column=custom_column)
        return _data, _pager


class RuntimeTermViewJson(RuntimeTermView):
    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        _data, _pager = super().list(project_uri, fullname, show_removed, page, size)
        cls.pretty_json(_data)

    def info(self, fullname: bool = False) -> None:
        _data = self.get_info_data(self.runtime.info(), fullname=fullname)
        self.pretty_json(_data)


def get_term_view(ctx_obj: t.Dict) -> t.Type[RuntimeTermView]:
    return (
        RuntimeTermViewJson if ctx_obj.get("output") == "json" else RuntimeTermViewRich
    )
